"""
Main engine for PyTorch
"""

from __future__ import annotations
from typing import Optional, Any, Union, Callable, Dict, Set
from contextlib import nullcontext, ExitStack, contextmanager

import gc
import os
import time
import socket
import fnmatch
import re
import math

import torch
import torch.distributed
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data import DataLoader
from torch import autocast
from torch.cuda import amp
import numpy as np

import returnn
from returnn.config import Config
from returnn.log import log
from returnn.engine.base import EngineBase
import returnn.frontend as rf
from returnn.tensor import TensorDict, Tensor, Dim
from returnn.datasets.basic import init_dataset, Dataset
from returnn.util import basic as util
from returnn.util import NumbersDict
from returnn.util.basic import hms, NotSpecified
from returnn.util.result_with_reason import ResultWithReason
from returnn.util.debug import debug_shell
from returnn.util.math import simplify_and_format_number, merge_random_seeds
from returnn.forward_iface import ForwardCallbackIface

from .updater import Updater
from .data import pipeline as data_pipeline
from .data import returnn_dataset_wrapper
from .data import extern_data as extern_data_util
from .data.queued_data_iter import QueuedDataIter
from .frontend.bridge import rf_module_to_pt_module
from .util import diagnose_gpu
from .util import module as util_module
from .util.exception_helper import help_on_torch_exception
from .util.debug_inf_nan import debug_inf_nan
from .distributed import DistributedContext, get_ctx as dist_get_ctx


class Engine(EngineBase):
    """
    PyTorch engine
    """

    def __init__(self, config: Config):
        """
        :param config:
        """
        super(Engine, self).__init__(config=config)
        rf.select_backend_torch()
        if util.BackendEngine.selected_engine is None:
            util.BackendEngine.select_engine(default_fallback_engine=util.BackendEngine.Torch, config=self.config)
        self.model_filename = self.config.value("model", None)
        self._mp_manager = torch.multiprocessing.Manager()
        self._epoch_mp_shared = self._mp_manager.Value("i", 0)
        self.train_dataset = None  # type: Optional[Dataset]
        self.eval_datasets = {}
        self.extern_data = None  # type: Optional[TensorDict]
        self._train_dataloader = None  # type: Optional[DataLoader]
        self._eval_dataloaders = {}  # type: Dict[str, DataLoader]

        self._start_epoch = None  # type: Optional[int]
        self._final_epoch = None  # type: Optional[int]
        self._min_seq_length = config.typed_value("min_seq_length", None) or config.int(
            "min_seq_length", None
        )  # type: Union[int,float,Dict[str,int],NumbersDict]
        self._max_seq_length = config.typed_value("max_seq_length", None) or config.int(
            "max_seq_length", None
        )  # type: Union[int,float,Dict[str,int],NumbersDict]
        self._orig_model = None  # type: Optional[Union[rf.Module, torch.nn.Module]]
        self._pt_model = None  # type: Optional[torch.nn.Module]
        self._train_step_func = None  # type: Optional[Callable]
        self._forward_step_func = self.config.typed_value("forward_step")  # type: Optional[Callable]
        self._forward_step_expected_outputs = None  # type: Optional[TensorDict]
        if self.config.typed_value("model_outputs") is not None:
            self._forward_step_expected_outputs = TensorDict()
            self._forward_step_expected_outputs.update(self.config.typed_value("model_outputs"), auto_convert=True)
        self._save_model_epoch_interval = 1
        self._ignore_param_set: Set[str] = set()  # for the updater and for saving the model checkpoint
        self._updater = None  # type: Optional[Updater]

        self._use_autocast = False
        self._autocast_dtype = None  # type: Optional[str]
        self._grad_scaler = None  # type: Optional[amp.GradScaler]

        dev_ = get_device_from_config_opt(config.value("device", None))
        self._device = dev_.result
        print("Using device:", self._device, f"({dev_.reason or '?'})", file=log.v2)

        self._torch_distributed_ctx = None  # type: Optional[DistributedContext]
        self._ddp_pt_model = None  # type: Optional[DistributedDataParallel]

        if config.typed_value("torch_distributed") is not None:
            self._torch_distributed_ctx = dist_get_ctx(config=config)
            local_rank = self._torch_distributed_ctx.local_rank()
            print(f"Start running torch distributed training on local rank {local_rank}.", file=log.v2)
            assert self._device == "cuda", f"torch distributed: unexpected device {self._device!r}"
            self._device = f"cuda:{local_rank}"

        if self._device == "cuda" or self._device.startswith("cuda:"):
            # Theano and TensorFlow print sth like: Using gpu device 2: GeForce GTX 980 (...)
            # Print in a similar format so that some scripts which grep our stdout work just as before.
            diagnose_gpu.print_using_cuda_device_report(self._device, file=log.v2)

        if self._device.startswith("cuda:"):
            # There is potentially lots of code which would use the default CUDA device,
            # even sometimes ignoring the input device.
            # For example, DistributedDataParallel _verify_param_shape_across_processes does that
            # (https://github.com/rwth-i6/returnn/issues/1469 https://github.com/pytorch/pytorch/issues/114765).
            # The user code potentially also just uses "cuda" as device, which would use the default device.
            # Thus, to be on the safe side, we set the default device to the one we want to use.
            torch.cuda.set_device(self._device)

        self._log_memory_usage = config.bool("torch_log_memory_usage", False)
        self._log_batch_size = config.bool("log_batch_size", False) and log.verbose[5]
        self._calculate_exp_loss = config.bool("calculate_exp_loss", False)
        self._log_grad_norm = _parse_log_grad_norm(config)
        self._reset_dev_memory_caches = config.bool("reset_dev_memory_caches", False)
        self._forward_auto_split_batch_on_oom = config.bool("forward_auto_split_batch_on_oom", False)
        self._stop_on_nonfinite_train_score = config.bool("stop_on_nonfinite_train_score", True)

        default_float_dtype = config.value("default_float_dtype", None)
        if default_float_dtype is not None:
            assert isinstance(default_float_dtype, str)
            default_float_dtype = getattr(torch, default_float_dtype)
            assert isinstance(default_float_dtype, torch.dtype)
        self._default_float_dtype: Optional[torch.dtype] = default_float_dtype

        amp_options = self.config.opt_typed_value("torch_amp")
        grad_scaler_opts = self.config.typed_value("grad_scaler", NotSpecified)
        if amp_options is not None:
            self._use_autocast = True
            if isinstance(amp_options, dict):
                amp_options = util.CollectionReadCheckCovered(amp_options)
                dtype = amp_options.get("dtype", None)
                grad_scaler_opts = amp_options.get("grad_scaler", grad_scaler_opts)
                amp_options.assert_all_read()
            elif isinstance(amp_options, str):
                dtype = amp_options
            else:
                raise TypeError(f"Invalid type for torch_amp: {type(amp_options)}")
            if isinstance(dtype, str):
                dtype = getattr(torch, dtype)
            assert isinstance(dtype, torch.dtype) or dtype is None
            print(f"Using autocast (automatic mixed precision (AMP)) with dtype {dtype}", file=log.v2)
            self._autocast_dtype = dtype

        if grad_scaler_opts is NotSpecified:
            grad_scaler_opts = {} if self._use_autocast else None
        if grad_scaler_opts is not None:
            assert isinstance(grad_scaler_opts, dict)
            print("Using GradScaler with options:", grad_scaler_opts, file=log.v2)
            self._grad_scaler = amp.GradScaler(**grad_scaler_opts)

    def init_network_from_config(self, config: Optional[Config] = None):
        """init model"""
        assert config is self.config or not config
        super().init_network_from_config(config=config)

        extern_data_dict = self.config.typed_value("extern_data")
        assert extern_data_dict, "extern_data is not specified in config"
        self.extern_data = extern_data_util.extern_data_template_from_config_opts(extern_data_dict)

        self._load_model()

    def init_train_from_config(
        self,
        config: Optional[Config] = None,
        train_data: Optional[Dataset] = None,
        dev_data: Optional[Dataset] = None,
        eval_data: Optional[Dataset] = None,
    ):
        """
        :param config:
        :param train_data:
        :param dev_data:
        :param eval_data:
        """
        assert config is self.config or not config
        config = self.config
        super().init_train_from_config(config=config)

        self.train_dataset = train_data
        self.eval_datasets.clear()
        if dev_data:
            self.eval_datasets["dev"] = dev_data
        if eval_data:
            self.eval_datasets["eval"] = eval_data
        if config.has("eval_datasets"):
            for dataset_name, dataset_opts in config.typed_value("eval_datasets", {}).items():
                self.eval_datasets[dataset_name] = init_dataset(dataset_opts, default_kwargs={"name": dataset_name})

        self._train_dataloader = self._create_data_loader(train_data, train=True) if train_data else None
        for dataset_name, dataset in self.eval_datasets.items():
            self._eval_dataloaders[dataset_name] = self._create_data_loader(dataset, train=False)

        self._start_epoch = self.get_train_start_epoch(self.config)
        self._final_epoch = self.config_get_final_epoch(self.config)

        self.init_network_from_config(config=config)

        self._save_model_epoch_interval = config.int("save_interval", 1)

        if self._torch_distributed_ctx:
            self._ddp_pt_model = self._torch_distributed_ctx.maybe_make_distributed_module(
                module=_WrappedModuleRunStep(module=self._pt_model, engine=self)
            )
        self._updater = Updater(
            config=self.config, network=self._pt_model, device=self._device, initial_learning_rate=self.learning_rate
        )
        self._updater.create_optimizer()
        if self._start_epoch > 1:
            self._load_optimizer(epoch=self._start_epoch - 1)

        self._train_step_func = self.config.typed_value("train_step")
        assert self._train_step_func, "train_step not defined"

    def set_epoch(self, epoch: int):
        """set epoch"""
        super().set_epoch(epoch)
        self._epoch_mp_shared.value = epoch

    def train(self):
        """
        Main training loop.
        """
        assert self._pt_model is not None, "Model not initialized, call init_train_from_config()."
        assert self.train_dataset is not None, "Train dataset missing, call init_train_from_config() with train_data."

        self._check_missing_eval()

        if self._start_epoch > self._final_epoch:
            print(f"Already trained until final epoch {self._final_epoch}, nothing to do.", file=log.v3)
            return

        print(
            f"Starting training at epoch {self._start_epoch}, global train step {self.global_train_step}", file=log.v3
        )
        self.set_epoch(self._start_epoch - 1)
        while self.epoch + 1 <= self._final_epoch:
            self.set_epoch(self.epoch + 1)
            self.init_train_epoch()
            self.train_epoch()

        print(f"Finished training at epoch {self.epoch}, global train step {self.global_train_step}", file=log.v3)

    def init_train_epoch(self):
        """
        init train (sub)epoch. LR etc
        """
        self.learning_rate = self.learning_rate_control.get_learning_rate_for_epoch(self.epoch)

        # Update learning rate
        self._updater.set_learning_rate(self.learning_rate)
        self._updater.set_current_train_step(
            global_train_step=self.global_train_step, epoch=self.epoch, epoch_continuous=self.epoch - 1
        )

        self.learning_rate_control.epoch_data[self.epoch].meta.update(
            {
                "global_train_step": self.global_train_step,
                "effective_learning_rate": self._updater.get_effective_learning_rate(),
                "returnn": util.describe_returnn_version(),
                "torch": util.describe_torch_version(),
                "time": time.strftime("%Y-%m-%d-%H-%M-%S (UTC%z)"),
                "hostname": socket.gethostname(),
                "device": torch.cuda.get_device_name() if torch.device(self._device).type == "cuda" else self._device,
                "cpu": util.get_cpu_model_name(),
                "distributed": self._torch_distributed_ctx.__repr__() if self._torch_distributed_ctx else None,
            }
        )

        # Note: The RF/Torch default random number generator influences many things during training,
        # such as dropout and other random operations inside the model,
        # but also some potential shuffling in the dataset iterator.
        # Also see Dataset._get_default_random_seed_offset() and Dataset._get_random_seed_for_epoch().
        random_seed = self.config.int("random_seed", 42)
        seed_data = [self.epoch, self.global_train_step, random_seed]
        if self._torch_distributed_ctx:
            seed_data.append(self._torch_distributed_ctx.rank())
        random_seed = merge_random_seeds(seed_data)  # Join all seeds into one int.
        rf.set_random_seed(random_seed)

    def _maybe_reset_dev_memory_caches(self, *, force: bool = False):
        if not force and not self._reset_dev_memory_caches:
            return
        gc.collect()
        torch.cuda.empty_cache()

    def _reset_dev_memory_stats(self):
        dev = torch.device(self._device)
        if dev.type == "cuda":
            torch.cuda.reset_peak_memory_stats(dev)
        self._maybe_report_dev_memory_stats()

    def _maybe_report_dev_memory_stats(self):
        if not self._log_memory_usage:
            return
        dev = torch.device(self._device)
        if dev.type == "cuda":
            stats = [
                f"alloc cur {util.human_bytes_size(torch.cuda.memory_allocated(dev))}",
                f"alloc peak {util.human_bytes_size(torch.cuda.max_memory_allocated(dev))}",
                f"reserved cur {util.human_bytes_size(torch.cuda.memory_reserved(dev))}",
                f"reserved peak {util.human_bytes_size(torch.cuda.max_memory_reserved(dev))}",
            ]
            print(f"Memory usage ({self._device}):", " ".join(stats), file=log.v1)

    def train_epoch(self):
        """
        train one (sub)epoch
        """
        print(
            "start",
            self.get_epoch_str(),
            "global train step",
            self.global_train_step,
            "with effective learning rate",
            self._updater.get_effective_learning_rate(),
            "...",
            file=log.v3,
        )

        accumulated_losses_dict = NumbersDict()
        accumulated_inv_norm_factors_dict = NumbersDict()
        step_idx = 0
        epoch_start_time = time.monotonic()

        data_iter = iter(self._train_dataloader)
        elapsed_computation_time = 0

        self._pt_model.train()
        self._maybe_reset_dev_memory_caches()
        self._reset_dev_memory_stats()

        if self.config.bool("debug_shell_before_train_loop", False):
            print("debug_shell_before_train_loop", file=log.v1)
            debug_shell(user_ns=locals(), user_global_ns=globals(), exit_afterwards=False)

        accum_grad_multiple_step_dyn = None
        if self.config.typed_dict.get("accum_grad_multiple_step") is not None:
            v = self.config.typed_dict["accum_grad_multiple_step"]
            if isinstance(v, int):
                accum_grad_multiple_step = v
            else:
                assert callable(v), "accum_grad_multiple_step should be an int or callable"
                accum_grad_multiple_step = 0
                accum_grad_multiple_step_dyn = v
        else:
            accum_grad_multiple_step = self.config.int("accum_grad_multiple_step", 1)

        zero_grad_next_step = True
        cur_count_grad_accum = 0
        extern_data = None

        total_data_size_packed = NumbersDict()
        total_data_size_padded = NumbersDict()

        report_prefix = f"ep {self.epoch} train"
        try:
            while True:
                with torch.no_grad():
                    extern_data_raw = next(data_iter, None)

                step_begin_time = time.monotonic()

                if step_idx == 0 and log.verbose[5]:
                    print("Time to get first batch data:", hms(step_begin_time - epoch_start_time), file=log.v5)

                _has_data = torch.tensor([extern_data_raw is not None], dtype=torch.int8)
                if self._torch_distributed_ctx:
                    # use all reduce to check if all workers have data, if at least one worker does not have data,
                    # all workers finish this epoch
                    torch.distributed.all_reduce(_has_data, op=torch.distributed.ReduceOp.MIN)
                if not _has_data[0]:
                    break

                # convert values from torch int32 to Python ints to prevent overflow
                keys_w_seq_len = [k for k in extern_data_raw if f"{k}:seq_len" in extern_data_raw]
                total_data_size_packed += NumbersDict(
                    {k: int(sum(extern_data_raw[f"{k}:seq_len"])) for k in keys_w_seq_len},
                )
                total_data_size_padded += NumbersDict(
                    {k: int(util.prod(extern_data_raw[k].shape[:2])) for k in keys_w_seq_len},
                )

                complete_frac = float(extern_data_raw["complete_frac"])
                epoch_continuous = self.epoch - 1 + complete_frac if complete_frac >= 0.0 else None
                num_seqs = int(extern_data_raw["num_seqs"])

                # clear the gradients when every gradient accumulation loop starts
                if zero_grad_next_step:
                    self._updater.get_optimizer().zero_grad()
                    cur_count_grad_accum = 0

                extern_data = extern_data_util.raw_dict_to_extern_data(
                    extern_data_raw,
                    extern_data_template=self.extern_data,
                    device=self._device,
                    float_dtype=self._default_float_dtype,
                )
                self._run_step(extern_data, train_flag=True, train_func=True)

                train_ctx = rf.get_run_ctx()
                total_loss = train_ctx.total_loss()
                losses_dict = NumbersDict(
                    {
                        name: (
                            float(loss.get_summed_loss().raw_tensor.detach().cpu().item())
                            if self._device != "meta"
                            else float("nan")
                        )
                        for name, loss in train_ctx.losses.items()
                    }
                )
                inv_norm_factors_dict = NumbersDict(
                    {name: float(_to_raw(loss.get_inv_norm_factor())) for name, loss in train_ctx.losses.items()}
                )

                if accum_grad_multiple_step_dyn:
                    accum_grad_multiple_step = accum_grad_multiple_step_dyn(
                        epoch=self.epoch,
                        epoch_continuous=epoch_continuous,
                        global_train_step=self.global_train_step,
                        **util.get_fwd_compat_kwargs(),
                    )
                cur_count_grad_accum += 1
                perform_update_step = cur_count_grad_accum >= accum_grad_multiple_step
                with (
                    self._ddp_pt_model.no_sync()
                    if (self._ddp_pt_model is not None and not perform_update_step)
                    else nullcontext()
                ):
                    if self._grad_scaler is not None:
                        self._grad_scaler.scale(total_loss.raw_tensor).backward()
                    else:
                        total_loss.raw_tensor.backward()

                if self._log_grad_norm and perform_update_step:
                    key = f"grad_norm:p{simplify_and_format_number(self._log_grad_norm)}"
                    assert key not in losses_dict
                    inv_norm_factors_dict[key] = 1.0  # once per update step
                    losses_dict[key] = _get_total_grad_norm(self._pt_model, p=self._log_grad_norm)

                # only update the weights when every gradient accumulation loop ends
                if perform_update_step:
                    self._updater.step(grad_scaler=self._grad_scaler)
                zero_grad_next_step = perform_update_step

                if self._torch_distributed_ctx:
                    self._torch_distributed_ctx.step_after_param_update(module=self._pt_model, epoch_step_idx=step_idx)

                step_end_time = time.monotonic()
                step_duration = step_end_time - step_begin_time
                elapsed_computation_time += step_duration

                accumulated_losses_dict += losses_dict
                accumulated_inv_norm_factors_dict += inv_norm_factors_dict
                eval_info = self._maybe_extend_losses_info(losses_dict / inv_norm_factors_dict)
                _print_process(
                    report_prefix,
                    step=step_idx,
                    eval_info=dict(eval_info),
                    step_duration=step_duration,
                    start_elapsed=step_end_time - epoch_start_time,
                    complete_frac=complete_frac,
                    num_seqs=num_seqs,
                    batch_size_info=_get_batch_size_info(extern_data) if self._log_batch_size else None,
                    log_memory_usage_device=self._device if self._log_memory_usage else None,
                )

                if self._stop_on_nonfinite_train_score:
                    if any(np.isinf(v) or np.isnan(v) for v in accumulated_losses_dict.values()):
                        print("Model seems broken, got inf or nan score.", file=log.v1)
                        print(
                            "Accumulated scores:",
                            accumulated_losses_dict / accumulated_inv_norm_factors_dict,
                            file=log.v1,
                        )

                        print("Checking for inf/nan in model parameters...", file=log.v1)
                        count_nan_inf_params = 0
                        for name, param in self._pt_model.named_parameters():
                            got_nan_inf_t = torch.stack([torch.isnan(param).any(), torch.isinf(param).any()]).cpu()
                            got_nan = got_nan_inf_t[0].item()
                            got_inf = got_nan_inf_t[1].item()
                            if got_nan or got_inf:
                                s = "/".join([s_ for s_, b in [("nan", got_nan), ("inf", got_inf)] if b])
                                print(f"  {name} {param}: {s}", file=log.v1)
                                count_nan_inf_params += 1
                        if count_nan_inf_params == 0:
                            print("(No inf/nan in model parameters.)", file=log.v1)

                        def _debug_func() -> torch.Tensor:
                            self._run_step(extern_data, train_flag=True, train_func=True)
                            loss = rf.get_run_ctx().total_loss()
                            assert isinstance(loss, Tensor)
                            return loss.raw_tensor

                        print("Running debug_inf_nan...", file=log.v1)
                        debug_inf_nan(_debug_func, with_grad=True)
                        if count_nan_inf_params > 0 and self.global_train_step == 1:
                            print(
                                "This was the second step, so likely the first step grad was broken."
                                " Try again with reset model...",
                                file=log.v1,
                            )
                            self._load_model()
                            debug_inf_nan(_debug_func, with_grad=True)
                        raise Exception(f"Inf/nan score in step {step_idx}.")

                step_idx += 1
                self.global_train_step += 1
                self._updater.set_current_train_step(
                    global_train_step=self.global_train_step, epoch=self.epoch, epoch_continuous=epoch_continuous
                )
        except Exception as exc:
            help_on_torch_exception(exc, step_idx=step_idx, model=self._orig_model, extern_data=extern_data)
            raise

        elapsed = time.monotonic() - epoch_start_time
        elapsed_computation_percentage = elapsed_computation_time / elapsed
        total_padding_ratio = NumbersDict.constant_like(1.0, total_data_size_packed) - (
            total_data_size_packed / total_data_size_padded
        )
        assert 0.0 <= total_padding_ratio.min_value() <= total_padding_ratio.max_value() <= 1.0
        pad_str = ", ".join(f"{k}: {v:.1%}" for k, v in total_padding_ratio.items())
        print(
            f"Epoch {self.epoch}: Trained {step_idx} steps, {hms(elapsed)} elapsed "
            f"({elapsed_computation_percentage:.1%} computing time, {pad_str} padding)",
            file=log.v3,
        )

        self.learning_rate_control.epoch_data[self.epoch].meta.update(
            {
                "epoch_num_train_steps": step_idx,
                "epoch_train_time_secs": round(elapsed),
                "global_train_step_end": self.global_train_step,
            }
        )

        accumulated_losses_dict = accumulated_losses_dict / accumulated_inv_norm_factors_dict
        accumulated_losses_dict = self._maybe_extend_losses_info(accumulated_losses_dict)
        self.learning_rate_control.set_epoch_error(
            self.epoch, {f"train_loss_{k}": v for k, v in accumulated_losses_dict.items()}
        )
        if self._do_save():
            self.learning_rate_control.save()

        print(f"Epoch {self.epoch}: Total train loss:", _format_score(dict(accumulated_losses_dict)), file=log.v3)

        self._maybe_report_dev_memory_stats()

        if self.epoch % self._save_model_epoch_interval == 0 or self.epoch == self._final_epoch:
            if self.model_filename:
                self._save_model()
                self._save_optimizer()
            else:
                print("Not saving model, `model` not specified.", file=log.v3)

        self.eval_model()
        if self.config.bool_or_other("cleanup_old_models", None):
            self.cleanup_old_models()

    def _do_save(self):
        if self._device == "meta":
            return False
        if not super()._do_save():
            return False
        return True

    def eval_model(self, *, skip_already_evaluated: bool = False):
        """
        Runs model on all eval datasets and calculates the loss.
        """
        self._pt_model.eval()
        self._maybe_reset_dev_memory_caches()
        self._reset_dev_memory_stats()

        eval_dump_str = []

        for dataset_name, dataset in self.eval_datasets.items():
            if skip_already_evaluated and self._is_dataset_evaluated(name=dataset_name):
                continue

            data_loader = self._eval_dataloaders[dataset_name]

            if self._torch_distributed_ctx and self._torch_distributed_ctx.rank() != 0:
                # We need to make sure the data loader iterator was created for proper synchronization.
                # However, we only want to do evaluation on rank 0 for simplicity.
                iter(data_loader)
                # We wait here until rank 0 is done w/ the eval.
                # Regularly synchronizing should fix potential timeout issues, like
                # https://github.com/rwth-i6/returnn/issues/1621.
                _has_data = torch.tensor([True], device="cpu", dtype=torch.int8)
                while _has_data:
                    torch.distributed.broadcast(_has_data, src=0)
                continue

            print(f"Evaluating dataset {dataset_name!r}", file=log.v3)

            accumulated_losses_dict = NumbersDict()
            accumulated_inv_norm_factors_dict = NumbersDict()
            step_idx = 0
            eval_start_time = time.monotonic()

            report_prefix = f"ep {self.epoch} {dataset_name} eval"
            with torch.no_grad():
                for extern_data_raw in data_loader:
                    if self._torch_distributed_ctx and step_idx % 100 == 0:
                        _has_data = torch.tensor([True], device="cpu", dtype=torch.int8)
                        torch.distributed.broadcast(_has_data, src=0)

                    complete_frac = float(extern_data_raw["complete_frac"])
                    num_seqs = int(extern_data_raw["num_seqs"])

                    extern_data = extern_data_util.raw_dict_to_extern_data(
                        extern_data_raw,
                        extern_data_template=self.extern_data,
                        device=self._device,
                        float_dtype=self._default_float_dtype,
                    )

                    self._run_step(extern_data, train_func=True)
                    step_end_time = time.monotonic()

                    train_ctx = rf.get_run_ctx()

                    losses_dict = NumbersDict(
                        {
                            name: (
                                float(loss.get_summed_loss().raw_tensor.detach().cpu().item())
                                if self._device != "meta"
                                else float("nan")
                            )
                            for name, loss in train_ctx.losses.items()
                        }
                    )
                    inv_norm_factors_dict = NumbersDict(
                        {name: float(_to_raw(loss.get_inv_norm_factor())) for name, loss in train_ctx.losses.items()}
                    )

                    accumulated_losses_dict += losses_dict
                    accumulated_inv_norm_factors_dict += inv_norm_factors_dict
                    eval_info = self._maybe_extend_losses_info(losses_dict / inv_norm_factors_dict)
                    _print_process(
                        report_prefix,
                        step=step_idx,
                        eval_info=dict(eval_info),
                        complete_frac=complete_frac,
                        num_seqs=num_seqs,
                        start_elapsed=step_end_time - eval_start_time,
                        log_memory_usage_device=self._device if self._log_memory_usage else None,
                    )
                    step_idx += 1

            assert step_idx > 0, f"No data in dataset {dataset_name!r}."
            accumulated_losses_dict = accumulated_losses_dict / accumulated_inv_norm_factors_dict
            accumulated_losses_dict = self._maybe_extend_losses_info(accumulated_losses_dict)

            self.learning_rate_control.set_epoch_error(
                self.epoch, {f"{dataset_name}_loss_{k}": v for k, v in accumulated_losses_dict.items()}
            )
            if self._do_save():
                self.learning_rate_control.save()

            # Same format as the TF engine.
            eval_dump_str += ["%s: %s" % (dataset_name, _format_score(dict(accumulated_losses_dict)))]

            if self._torch_distributed_ctx:
                assert self._torch_distributed_ctx.rank() == 0
                _has_data = torch.tensor([False], device="cpu", dtype=torch.int8)
                torch.distributed.broadcast(_has_data, src=0)

        if not self._torch_distributed_ctx or self._torch_distributed_ctx.rank() == 0:
            print(
                f"Epoch {self.epoch} evaluation:",
                " ".join(eval_dump_str) if eval_dump_str else "(No evaluations.)",
                file=log.v1,
            )

        self._maybe_report_dev_memory_stats()

        # Maybe sync train/dev/eval scores for the epoch.
        if self._torch_distributed_ctx:
            ls = [self.learning_rate_control.epoch_data[self.epoch]]
            torch.distributed.broadcast_object_list(ls, src=0, device=torch.device("cpu"))
            assert isinstance(ls[0], self.learning_rate_control.EpochData)
            self.learning_rate_control.epoch_data[self.epoch] = ls[0]

    def _maybe_extend_losses_info(self, losses: NumbersDict) -> NumbersDict:
        """
        :param losses:
        :return: maybe extended losses
        """
        if self._calculate_exp_loss and losses.has_values():
            # Assume the current run ctx still has info about the losses from the last step.
            assert rf.get_run_ctx().losses
            score_keys = set(k for k, v in rf.get_run_ctx().losses.items() if not v.as_error)
            losses_ = {}
            for key, value in losses.items():
                losses_[key] = value
                if key in score_keys:
                    try:
                        losses_[f"{key}:exp"] = math.exp(value)
                    except OverflowError:
                        losses_[f"{key}:exp"] = float("inf")
            losses = NumbersDict(losses_)
        return losses

    def _create_data_loader(
        self, dataset: Dataset, *, train: bool = False, dataset_init_epoch: bool = True
    ) -> DataLoader:
        """
        :param dataset: RETURNN dataset
        :param train: Train might use a separate batch size in the config (batch_size_train vs batch_size_dev).
            Also online_shuffle_batches is only used in training.
        :param dataset_init_epoch: Whether to call dataset.init_seq_order(epoch=self.epoch) or not.
        :return: PyTorch data loader created from given RETURNN dataset
        """
        # Make sure that _dataset_reset does not keep a ref to `self`,
        # otherwise it would trigger to pickle `self` and all its members.
        if dataset_init_epoch:
            dataset_reset = returnn_dataset_wrapper.ReturnnDatasetResetMpSharedEpochCallback(
                dataset=dataset, epoch_mp_shared=self._epoch_mp_shared
            )
        else:
            dataset_reset = returnn_dataset_wrapper.ReturnnDatasetResetNoOpCallback()

        wrapped_dataset = returnn_dataset_wrapper.ReturnnDatasetIterDataPipe(dataset, reset_callback=dataset_reset)
        if (self._min_seq_length is not None) or (self._max_seq_length is not None):
            wrapped_dataset = data_pipeline.LenFilterDataPipe(
                wrapped_dataset, min_seq_length=self._min_seq_length, max_seq_length=self._max_seq_length
            )
        chunking = self.config.typed_value("chunking", None)
        min_chunk_size = self.config.typed_value("min_chunk_size", 0)
        if chunking:
            wrapped_dataset = data_pipeline.ChunkingIterDataPipe(
                wrapped_dataset, chunking, min_chunk_size=min_chunk_size
            )

        batches_dataset = data_pipeline.get_batching_iterable_dataset_from_config(
            dataset=wrapped_dataset, config=self.config, train=train
        )

        online_shuffle_batches = self.config.typed_value("online_shuffle_batches", None)
        if train and online_shuffle_batches:
            if isinstance(online_shuffle_batches, int):
                online_shuffle_batches = {"buffer_size": online_shuffle_batches}
            elif isinstance(online_shuffle_batches, dict):
                if "buffer_size" not in online_shuffle_batches:
                    raise ValueError(
                        f"config online_shuffle_batches, buffer_size not defined, got {online_shuffle_batches}"
                    )
            else:
                raise TypeError(
                    f"config online_shuffle_batches, expected int or dict, got {type(online_shuffle_batches)}"
                )
            # Note on random seed: This is handled by the PyTorch DataLoader iterator logic and IterDataPipe reset.
            # Specifically, when we create a new DataLoader iterator,
            # this will get fetch a new random number (from current Torch RNG state),
            # use that as seed for the shuffle buffer.
            # Note: In case of distributed training, it will broadcast the seed from rank 0 to all others.
            # This is maybe not really what we want?
            # https://discuss.pytorch.org/t/shuffleriterdatapipe-but-different-random-seed-per-distributed-rank/212612
            # I currently don't really see a good way to override this behavior.
            # Also note that we are likely using persistent multiprocessing data loader workers,
            # so calling torch.utils.data.graph_settings.apply_random_seed here in the main proc
            # will not have an effect then.
            batches_dataset = data_pipeline.ShufflingDataPipe(
                batches_dataset, monotonic_data_keys=("complete_frac", "seq_idx"), **online_shuffle_batches
            )

        loader_opts = self.config.typed_value("torch_dataloader_opts") or {}
        assert isinstance(loader_opts, dict), f"config torch_dataloader_opts, expected dict, got {type(loader_opts)}"

        if not dataset_init_epoch:
            loader_opts = loader_opts.copy()
            # We want to keep the current initialized seq order of the dataset.
            # Multiprocessing does not quite work with this
            # (the serialization of the dataset state does not cover the current seq order).
            loader_opts["num_workers"] = 0

        data_loader = data_pipeline.create_data_loader_from_batches(batches_dataset, loader_opts)

        if data_loader.num_workers > 0:  # uses multi processing
            # We are not using the dataset anymore here in the main proc,
            # so free all resources as much as we can.
            # https://github.com/rwth-i6/returnn/issues/1443
            # Do this after creating the data loader - so in case it used the fork start method,
            # it would still potentially have resources ready to use.
            dataset.finish_epoch(free_resources=True)

        return data_loader

    @contextmanager
    def _run_ctx_mgr(self):
        with ExitStack() as stack:
            if self._use_autocast:
                stack.enter_context(autocast(device_type=self._device.split(":")[0], dtype=self._autocast_dtype))
            stack.enter_context(rf.set_default_device_ctx(self._device))
            if self._default_float_dtype:
                stack.enter_context(rf.set_default_float_dtype_ctx(str(self._default_float_dtype).split(".")[-1]))
                stack.enter_context(_set_torch_default_dtype_ctx_mgr(self._default_float_dtype))
            yield

    def _run_step(
        self, extern_data: TensorDict, *, train_flag: bool = False, train_func: bool, _inside_wrapped: bool = False
    ):
        """
        :param extern_data: model inputs for the step
        :return: Nothing, all outputs are written to the run context (:func:`rf.get_run_ctx`).
        """
        if self._ddp_pt_model is not None and not _inside_wrapped:
            self._ddp_pt_model(extern_data=extern_data, train_flag=train_flag, train_func=train_func)
            return

        if train_func:
            assert self._train_step_func is not None
            rf.init_train_step_run_ctx(train_flag=train_flag, step=self.global_train_step, epoch=self.epoch)
        else:
            assert self._forward_step_func is not None, "define forward_step in the config"
            rf.init_forward_step_run_ctx(
                expected_outputs=self._forward_step_expected_outputs, step=self.global_train_step, epoch=self.epoch
            )

        with self._run_ctx_mgr():
            sentinel_kw = util.get_fwd_compat_kwargs()
            if train_func:
                self._train_step_func(model=self._orig_model, extern_data=extern_data, **sentinel_kw)
            else:
                self._forward_step_func(model=self._orig_model, extern_data=extern_data, **sentinel_kw)

    def _load_model(self):
        """
        Sets self._model to a torch.nn.Module.
        """
        # Check existing model. This takes `load` and `load_epoch` into account,
        # and also whether we are in train or eval mode.
        epoch, model_epoch_filename = self.get_epoch_model(self.config)
        step = None

        filename = None
        checkpoint_state = None
        if model_epoch_filename:
            filename = model_epoch_filename + util.get_model_filename_postfix()
            if not os.path.exists(filename) and os.path.exists(model_epoch_filename):
                filename = model_epoch_filename
            print("Load model %s" % (filename,), file=log.v4)
            checkpoint_state = torch.load(filename, map_location=self._device)
            if epoch is None:
                epoch = checkpoint_state.get("epoch", self._start_epoch or 1)
            step = checkpoint_state.get("step", 1)
            print(f"  epoch {epoch}, global train step {step}", file=log.v4)
            # The checkpoint was saved when the step was already increased (but not the epoch yet).
            # Restore the last step.
            # Below, we will increase the step again in case we are training.
            step -= 1

        is_training = self.config.value("task", "train") == "train"
        is_first_train_epoch = not epoch and (is_training or self.config.value("task", "train") == "initialize_model")

        if not model_epoch_filename:
            step = 0
            epoch = self._start_epoch or 1

        self._create_model(epoch=epoch, step=step)

        non_critical_for_restore_set = set()
        if isinstance(self._orig_model, rf.Module):
            for key, param in self._orig_model.named_parameters():
                assert isinstance(param, rf.Parameter)
                if param.non_critical_for_restore:
                    non_critical_for_restore_set.add(key)

        self._ignore_param_set.clear()
        loaded_state_keys = set()
        missing_keys = set()
        unexpected_keys = set()
        if checkpoint_state is not None:
            model_state = checkpoint_state.get("model", checkpoint_state)
            loaded_state_keys.update(model_state.keys())
            missing_keys_main_ckpt, unexpected_keys_main_ckpt = self._pt_model.load_state_dict(
                model_state, strict=False
            )
            missing_keys.update(missing_keys_main_ckpt)
            unexpected_keys.update(unexpected_keys_main_ckpt)
            if unexpected_keys_main_ckpt:
                print(
                    f"Note: While loading {filename}, unexpected key(s) in state_dict: "
                    + ", ".join(map(repr, sorted(unexpected_keys_main_ckpt))),
                    file=log.v4,
                )
            del model_state
        # https://github.com/rwth-i6/returnn/issues/1345
        del checkpoint_state
        gc.collect()

        preload_from_files = self.config.typed_value("preload_from_files", {})
        if preload_from_files:
            model_state_keys_set = set(self._pt_model.state_dict().keys())
            # see `preload_from_files` in tf engine and `returnn.tf.network.CustomCheckpointLoader`
            # We use the reversed sorted order here to achieve consistent behavior with the TF engine.
            # There, the keys are used in sorted order but if a variable is loaded,
            # it will not be considered anymore afterward.
            # So the first occurrence is used.
            # Here, we overwrite variables even if they have been loaded before.
            # In order to get consistent behavior, we use the reversed order.
            for preload_key, opts in reversed(sorted(preload_from_files.items())):
                assert isinstance(opts, dict) and "filename" in opts
                init_for_train = opts.get("init_for_train", False)
                if init_for_train:
                    if isinstance(init_for_train, str) and init_for_train == "always":
                        # No matter if this is the first train epoch
                        # or training with loading some prev epoch,
                        # those parameters will always be loaded via preload_from_files,
                        # and thus also not stored in our own checkpoint.
                        pass
                    elif isinstance(init_for_train, bool) and init_for_train:
                        if not is_first_train_epoch:
                            continue
                    else:
                        raise ValueError(
                            f"preload key {preload_key}:"
                            f" invalid init_for_train value {init_for_train!r} (type {type(init_for_train).__name__})"
                        )
                else:  # default: init for recog
                    if is_training:
                        continue
                if opts["filename"] is None:
                    print(f"Pre-load (initialize) weights for key '{preload_key}'", file=log.v3)
                    if opts.get("pattern", None) is not None:
                        pattern = opts["pattern"]
                        match = re.compile(fnmatch.translate(pattern)).match
                    elif opts.get("prefix", None) is not None:
                        match = re.compile(re.escape(opts["prefix"]) + ".*").fullmatch
                    else:
                        raise ValueError(f"preload key {preload_key} without file {opts}: no pattern or prefix given")
                    remove = []
                    for name in self._pt_model.state_dict().keys():
                        if match(name) and name in missing_keys:
                            remove.append(name)
                    if remove:
                        print(f"Randomly initialize params: {remove}", file=log.v3)
                        missing_keys.difference_update(remove)
                    else:
                        print("(No relevant parameters matching.)", file=log.v3)
                    continue
                print(f"Pre-load weights for key '{preload_key}' from {opts['filename']}", file=log.v3)
                preload_model_state = torch.load(opts["filename"], map_location=self._device)
                if opts.get("checkpoint_key", "model") is not None:
                    # This can be used if an external checkpoint saves a checkpoint a different structure that just the
                    # model state dict. E.g., if a checkpoint is created using
                    # `torch.save({"model": model.state_dict(), "optimizer": optimizer.state)_dict(), ...})`
                    # we can set checkpoint_key = "model" to load the model.
                    # Currently, this only supports single level dicts, but it could be extended if needed.
                    preload_model_state = preload_model_state[opts.get("checkpoint_key", "model")]
                if opts.get("prefix", ""):
                    # Only params with this prefix should be loaded.
                    # They are expected to be in the checkpoint without this prefix.
                    # By adding the prefix to all params,
                    # we make sure that only params matching this condition are loaded.
                    # This is in line with the behavior of the TF engine.
                    preload_model_state = {opts["prefix"] + key: value for key, value in preload_model_state.items()}
                ignore_params = opts.get("ignore_params", [])
                ignore_params_prefixes = opts.get("ignore_params_prefixes", [])
                for key in list(preload_model_state.keys()):
                    if key in ignore_params or any(
                        [key.startswith(ignore_key) for ignore_key in ignore_params_prefixes]
                    ):
                        print(f"Ignoring variable {key}", file=log.v3)
                        preload_model_state.pop(key)
                for new_name, name_in_checkpoint in opts.get("var_name_mapping", {}).items():
                    preload_model_state[new_name] = preload_model_state.pop(name_in_checkpoint)
                if init_for_train == "always":
                    self._ignore_param_set.update(set(preload_model_state.keys()).intersection(model_state_keys_set))
                missing_keys_preload, unexpected_keys_preload = self._pt_model.load_state_dict(
                    preload_model_state, strict=False
                )
                preload_model_state_keys = set(preload_model_state.keys())
                loaded_state_keys.update(preload_model_state.keys())
                missing_keys.difference_update(preload_model_state.keys())
                del preload_model_state
                gc.collect()

                if opts.get("prefix", ""):
                    prefix_keys = [key for key in self._pt_model.state_dict() if key.startswith(opts.get("prefix", ""))]
                    if not prefix_keys:
                        raise Exception(
                            "No keys with prefix %r found in model.\nModel params:\n%s"
                            % (opts.get("prefix", ""), ", ".join(name for name, _ in self._pt_model.named_parameters()))
                        )
                else:
                    prefix_keys = model_state_keys_set
                missing_keys_preload = (
                    set(prefix_keys).intersection(set(missing_keys_preload)).difference(loaded_state_keys)
                )
                unexpected_keys_preload = (
                    set(prefix_keys).intersection(set(unexpected_keys_preload)).difference(loaded_state_keys)
                )
                if not preload_model_state_keys.intersection(prefix_keys):
                    raise Exception(
                        f"No keys with prefix {opts.get('prefix', '')!r} found in preload model state.\n"
                        f"Preload model state keys: {preload_model_state_keys}\n"
                        f"Model state keys: {model_state_keys_set}"
                    )
                if missing_keys_preload and not opts.get("ignore_missing", False):
                    missing_keys.update(missing_keys_preload)
                if missing_keys_preload:
                    print(f"Missing keys: {missing_keys_preload}", file=log.v4)
                if unexpected_keys_preload:
                    print(
                        f"Note: While loading preload_from_files {opts['filename']}, unexpected key(s) in state_dict: "
                        + ", ".join(map(repr, sorted(unexpected_keys_preload))),
                        file=log.v4,
                    )
                    unexpected_keys.update(unexpected_keys_preload)

        if self._ignore_param_set:
            util_module.convert_parameters_to_buffers(self._pt_model, self._ignore_param_set, persistent=False)

        if missing_keys.intersection(non_critical_for_restore_set):
            # We ignore missing keys which are non-critical for restore.
            print(
                "Ignoring missing keys which are non-critical for restore:",
                missing_keys.intersection(non_critical_for_restore_set),
                file=log.v4,
            )
            missing_keys.difference_update(non_critical_for_restore_set)
        if missing_keys:
            raise Exception(
                "\n".join(
                    [
                        f"While loading model {filename}, preload_from_files {bool(preload_from_files)}:",
                        "Unexpected key(s) in state_dict: " + ", ".join(map(repr, sorted(unexpected_keys))),
                        "Missing key(s) in state_dict: " + ", ".join(map(repr, sorted(missing_keys))),
                        "Any missing key is an error.",
                    ]
                )
            )

        if self._default_float_dtype:
            self._pt_model.to(dtype=self._default_float_dtype)
        self._pt_model.to(self._device)

        if model_epoch_filename and is_training:
            # We loaded a model from a checkpoint, and the epoch and step were taken from it.
            # The epoch was just finished, so we need to increment it.
            # We decremented the step above.
            epoch += 1
            step += 1
        self.set_epoch(epoch)  # in training, this will be reset to start_epoch
        self.global_train_step = step

        load_model_post_hooks = self.config.typed_value("load_model_post_hooks")
        if load_model_post_hooks:
            with self._run_ctx_mgr():
                sentinel_kw = util.get_fwd_compat_kwargs()
                for hook in load_model_post_hooks:
                    hook(model=self._orig_model, **sentinel_kw)

    def _create_model(self, *, epoch: int, step: int):
        """
        Set up self._pt_model and self._orig_model
        by calling get_model from the config.

        Note on the `epoch` and `step` args:
        In case we are loading a model:
          This is the epoch and step of the model we are loading.
        In case we are initializing a model:
          Epoch starts at 1, step starts at 0.
        The step is the global train step, i.e. the number of train steps we have done so far over all epochs,
        so it does not reset to 0 at each epoch.
        In a checkpoint, we stored the epoch of the most recent epoch we just finished.
        We stored the global train step after we already incremented it (that's why you have step -= 1 above).
        The checkpoint is always stored when we just have finished the epoch.

        :param epoch:
        :param step:
        """
        # See :mod:`rf.rand` docstring for an explanation of this logic.
        random_seed = self.config.int("random_seed", 42)
        random_seed = (epoch * 193939 + step * 19937 + random_seed * 27644437 + 479001599) % (2**31)
        rf.set_random_seed(random_seed)

        get_model_func = self.config.typed_value("get_model")
        assert get_model_func, "get_model not defined in config"
        sentinel_kw = util.get_fwd_compat_kwargs()
        model = get_model_func(epoch=epoch, step=step, device=self._device, **sentinel_kw)
        self._orig_model = model
        if isinstance(model, rf.Module):
            self._pt_model = rf_module_to_pt_module(model)
        elif isinstance(model, torch.nn.Module):
            self._pt_model = model
        else:
            raise TypeError(f"get_model returned {model} of type {type(model)}, expected rf.Module or torch.nn.Module")
        assert isinstance(self._pt_model, torch.nn.Module)
        print("Model:", self._pt_model, file=log.v4)
        num_params = sum([parameter.numel() for parameter in self._pt_model.parameters()])
        print(f"net params #: {num_params}", file=log.v2)

    def get_pt_model(self) -> Optional[torch.nn.Module]:
        """
        :return: PyTorch Module. in case this is using RF, it will return the wrapped module
        """
        return self._pt_model

    def _save_model(self):
        """
        Saves the state of self._model to file.
        """
        if not self._do_save():
            return
        filename = self.get_epoch_model_filename() + util.get_model_filename_postfix()
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        print("Save model under %s" % (filename,), file=log.v4)
        # First write to a temp-file, to be sure that writing happens without errors,
        # and only afterward rename to the target file.
        tmp_filename = filename + ".tmp_write"
        if os.path.exists(tmp_filename):
            os.unlink(tmp_filename)
        state_dict = self._pt_model.state_dict()
        if self._ignore_param_set:
            # Do some extra check that we don't save the ignored parameters.
            # Should not be in the state_dict anymore because we should have converted them to buffers
            # via util_module.convert_parameters_to_buffers before.
            remaining = set(state_dict.keys()).intersection(self._ignore_param_set)
            assert not remaining, f"_save_model: found remaining params in state_dict to ignore: {remaining}"
        torch.save(
            {
                "model": state_dict,
                "epoch": self.epoch,
                "step": self.global_train_step,
                "effective_learning_rate": self._updater.get_effective_learning_rate() if self._updater else None,
                "returnn_version": returnn.__long_version__,
            },
            tmp_filename,
        )
        os.rename(tmp_filename, filename)

    def get_pt_optimizer(self) -> Optional[torch.optim.Optimizer]:
        """
        :return: PyTorch optimizer
        """
        if not self._updater:
            return None
        return self._updater.get_optimizer()

    def _load_optimizer(self, *, epoch: int):
        """
        Loads a torch.optim.Optimizer from disk and uses it as the optimizer.
        This function is a wrapper to Updater.load_optimizer().
        """
        filename = self.get_epoch_model_filename(epoch=epoch) + ".opt" + util.get_model_filename_postfix()
        self._updater.load_optimizer(filename)

    def _save_optimizer(self):
        """
        Saves the optimizer state to a file.
        This function is a wrapper to Updater.save_optimizer().
        """
        if not self._do_save():
            return
        filename = self.get_epoch_model_filename() + ".opt" + util.get_model_filename_postfix()
        self._updater.save_optimizer(filename)

        # keep only the last two optimizer states (two in case one file gets corrupted)
        clean_epoch = self.epoch - 2
        apply_keep_logic = self.config.bool("apply_cleanup_old_models_to_optim_states", False)
        if clean_epoch > 0 and not apply_keep_logic:
            filename = self.get_epoch_model_filename(epoch=clean_epoch) + ".opt" + util.get_model_filename_postfix()
            if os.path.isfile(filename):
                os.unlink(filename)

    def forward_with_callback(
        self,
        *,
        dataset: Dataset,
        callback: ForwardCallbackIface,
        dataset_init_epoch: bool = True,
        allow_skipping_seqs: bool = False,
    ):
        """forward"""
        assert isinstance(dataset, Dataset)
        assert isinstance(callback, ForwardCallbackIface)

        epoch_start_time = time.monotonic()
        elapsed_computation_time = 0.0

        self._pt_model.eval()
        self._maybe_reset_dev_memory_caches()
        self._reset_dev_memory_stats()

        if dataset_init_epoch:
            if not self.config.bool("sort_dataset", True):
                pass
            elif dataset.seq_ordering == "sorted_reverse":
                pass
            elif dataset.supports_seq_order_sorting():
                # We can sort it. Sort it in reverse to make sure that we have enough memory right at the beginning.
                print("Dataset supports sorting, i.e. it will be sorted for optimal performance.", file=log.v3)
                dataset.seq_ordering = "sorted_reverse"
            else:
                print(
                    "Dataset does not support sorting, i.e. it will not be sorted for optimal performance.",
                    file=log.v3,
                )

        if allow_skipping_seqs:
            # Dangerous! If you enable this, you could lose sequences,
            # and your evaluation pipeline may silently produce incorrect results!
            print(
                f"Note: allow_skipping_seqs is enabled (with min_seq_length {self._min_seq_length},"
                f" max_seq_length {self._max_seq_length}),"
                f" this may lead to incorrect evaluation results!",
                file=log.v2,
            )
        else:
            assert (self._min_seq_length is None) and (self._max_seq_length is None), (
                f"min_seq_length {self._min_seq_length}, max_seq_length {self._max_seq_length} not allowed,"
                f" we want to keep all source sentences."
            )

        data_loader = self._create_data_loader(dataset, dataset_init_epoch=dataset_init_epoch)
        if self._forward_auto_split_batch_on_oom:
            data_loader = QueuedDataIter(data_loader)

        batch_dim = extern_data_util.get_batch_dim_from_extern_data(self.extern_data)

        def _get_tensor_wo_batch_numpy(x: Tensor) -> Tensor:
            if batch_dim not in x.dims:
                raise Exception(f"Expected {batch_dim} in {x}.")
            if x.dims.index(batch_dim) != 0:
                x = x.copy_move_axis(x.dims.index(batch_dim), 0)

            y_kwargs = x.copy_template_excluding_axis(0).get_kwargs()
            y_kwargs["dims"] = [_get_dim_tag_wo_batch(dim) for dim in y_kwargs["dims"]]
            y = Tensor(**y_kwargs)

            if x.batch_ndim > 1:
                raw = x.raw_tensor[batch_idx]
            else:
                # Keep it as ndarray.
                raw = x.raw_tensor[batch_idx : batch_idx + 1].reshape(())
            if any(d is not d_ for d, d_ in zip(x.dims[1:], y.dims)):  # replaced any dims?
                # Cut off any padding.
                raw = raw[tuple(slice(None, dim.get_dim_value()) for dim in y.dims)]
            # Convert it to Numpy array.
            # Note that users might also want to get the PyTorch tensors instead.
            # They might even want to get the whole batched tensor.
            # If that use cases comes up at some later point,
            # we can introduce it as an option, sth like "forward_raw_tensors" or so.
            # Currently, this callback interface is intended to also be used by other backends,
            # and then the user can always assume Numpy arrays.
            if isinstance(raw, torch.Tensor):  # might already be numpy array
                raw = raw.detach().cpu()
                if raw.dtype == torch.bfloat16:
                    raw = raw.float()
                raw = raw.numpy()
            y.raw_tensor = raw
            return y

        def _get_dim_tag_wo_batch(dim: Dim) -> Dim:
            """
            This is for dim tags with dyn_size_ext which include the batch_dim,
            e.g. the standard [batch] sizes.
            In the callback, we pass each sequence without the batch dim,
            so we must adapt the dim tags.
            """
            if dim.dyn_size_ext is None:
                return dim
            if batch_dim not in dim.dyn_size_ext.dims:
                return dim
            new_dim = dim.copy()
            new_dim.dyn_size_ext = _get_tensor_wo_batch_numpy(dim.dyn_size_ext)
            return new_dim

        report_prefix = f"ep {self.epoch} {dataset.name} forward"
        with torch.no_grad():
            callback.init(model=self._orig_model)

            step_idx = 0
            for extern_data_raw in data_loader:
                step_begin_time = time.monotonic()

                complete_frac = float(extern_data_raw["complete_frac"])
                num_seqs = int(extern_data_raw["num_seqs"])

                if self._forward_step_expected_outputs:
                    # Also resets any dyn dims, which might have been set in the prev step.
                    self._forward_step_expected_outputs.reset_content()
                extern_data = extern_data_util.raw_dict_to_extern_data(
                    extern_data_raw,
                    extern_data_template=self.extern_data,
                    device=self._device,
                    float_dtype=self._default_float_dtype,
                )
                try:
                    self._run_step(extern_data, train_func=False)
                except Exception as exc:
                    if (
                        isinstance(exc, torch.cuda.OutOfMemoryError)
                        and self._forward_auto_split_batch_on_oom
                        and extern_data_util.raw_dict_can_split_batch(extern_data_raw)
                    ):
                        help_on_torch_exception(exc, model=self._orig_model, always_direct_print=True)
                        util.traceback_clear_frames(exc.__traceback__)
                        diagnose_gpu.garbage_collect()
                        print(f"{report_prefix}, split step {step_idx} batch and try again...", file=log.v3)
                        data_loader.extend(extern_data_util.raw_dict_split_batch(extern_data_raw, splits=2))
                        continue
                    help_on_torch_exception(exc, model=self._orig_model)
                    raise
                ctx = rf.get_run_ctx()
                ctx.check_outputs_complete()

                model_outputs = ctx.outputs
                for batch_idx in range(batch_dim.get_dim_value()):
                    seq_tag = extern_data["seq_tag"].raw_tensor[batch_idx].item()
                    model_outputs_per_batch = TensorDict()
                    for k, v in model_outputs.data.items():
                        model_outputs_per_batch.data[k] = _get_tensor_wo_batch_numpy(v)
                    callback.process_seq(seq_tag=seq_tag, outputs=model_outputs_per_batch)

                step_end_time = time.monotonic()
                step_duration = step_end_time - step_begin_time
                elapsed_computation_time += step_duration

                _print_process(
                    report_prefix,
                    step=step_idx,
                    eval_info=None,
                    step_duration=step_duration,
                    start_elapsed=step_end_time - epoch_start_time,
                    complete_frac=complete_frac,
                    num_seqs=num_seqs,
                    batch_size_info=_get_batch_size_info(extern_data) if self._log_batch_size else None,
                    log_memory_usage_device=self._device if self._log_memory_usage else None,
                )
                step_idx += 1

            callback.finish()

        elapsed = time.monotonic() - epoch_start_time
        elapsed_computation_percentage = elapsed_computation_time / elapsed
        print(
            "Forward %i steps, %s elapsed (%.1f%% computing time)"
            % (step_idx, hms(elapsed), (elapsed_computation_percentage * 100.0)),
            file=log.v3,
        )

        self._maybe_report_dev_memory_stats()

    @staticmethod
    def delete_model(filename):
        """
        :param str filename:
        :return: accumulated file-size in bytes of deleted files
        :rtype: int
        """
        # This assumes PyTorch models here.
        # They consist of a file with the extension ".pt" and potentially an optimizer state with extension ".opt.pt"

        count_bytes = 0
        fname = filename + ".pt"
        assert os.path.exists(fname)
        count_bytes += os.stat(fname).st_size
        os.remove(fname)
        opt_fname = filename + ".opt.pt"
        if os.path.exists(opt_fname):
            count_bytes += os.stat(opt_fname).st_size
            os.remove(opt_fname)
        assert count_bytes > 0
        return count_bytes

    def _check_missing_eval(self):
        """
        Checks if there are outstanding tasks (eval_model) for the epoch,
        and executes them.
        """
        if not self.learning_rate_control.filename:
            return
        if self._start_epoch == 1:
            return

        # If the learning rate (scores) file is somehow corrupt,
        # check that and stop, otherwise we would potentially corrupt the data even more.
        for epoch in range(self.get_start_epoch_no_existing_model(self.config), self._start_epoch - 1):
            for name in ["train"] + list(self.eval_datasets.keys()):
                if not self._is_dataset_evaluated(name=name, epoch=epoch):
                    raise Exception(f"Scores of epoch {epoch} for {name} are missing.")

        self.set_epoch(self._start_epoch - 1)
        if not self._is_dataset_evaluated(name="train"):
            raise Exception(f"Scores of last train epoch {self.epoch} are missing.")
        for name in self.eval_datasets.keys():
            if not self._is_dataset_evaluated(name=name):
                # This can happen when we have a previous model but did not test it yet.
                print(f"Last epoch model not yet evaluated on {name}. Doing that now.", file=log.v3)
                self.eval_model(skip_already_evaluated=True)
                break


def _to_raw(n: Union[int, float, Tensor]):
    if isinstance(n, (int, float)):
        return n
    if isinstance(n, Tensor):
        x = n.raw_tensor.detach().cpu()
        if x.dtype == torch.bfloat16:
            x = x.float()
        return x.numpy()
    raise TypeError(f"Unexpected {n} of type {type(n)}")


def _print_process(
    report_prefix: str,
    *,
    step: int,
    eval_info: Optional[Dict[str, Any]] = None,
    batch_size_info: Optional[Dict[str, Any]] = None,
    step_duration: Optional[float] = None,
    start_elapsed: Optional[float] = None,
    complete_frac: Optional[float] = None,
    num_seqs: Optional[int] = None,
    log_memory_usage_device: Optional[str] = None,
):
    """
    Similar but simplified from TF engine _print_process.

    :param report_prefix:
    :param step: for this epoch
    :param eval_info:
    :param batch_size_info:
    :param step_duration: time elapsed for this step (secs)
    :param start_elapsed: time elapsed since epoch start (secs)
    :param complete_frac: how much of the current epoch is already consumed
    :param num_seqs: total number of seqs this epoch
    :param log_memory_usage_device: if given, will log memory usage (peak allocated memory)
    :return: nothing, will be printed to log
    """
    if log.verbose[5]:  # report every minibatch
        if step == 0 and num_seqs is not None and num_seqs >= 0:
            print(f"{report_prefix} num_seqs: {num_seqs}", file=log.v5)
        info = [report_prefix, "step %i" % step]
        if eval_info:  # Such as score.
            info += ["%s %s" % (k, _format_score_value(v)) for k, v in eval_info.items()]
        if batch_size_info:
            info += ["%s %s" % (k, _format_score_value(v)) for k, v in batch_size_info.items()]
        if log_memory_usage_device:
            dev = torch.device(log_memory_usage_device)
            if dev.type == "cuda":
                info += [
                    f"mem_usage:{log_memory_usage_device} {util.human_bytes_size(torch.cuda.max_memory_allocated(dev))}"
                ]
        if step_duration is not None:
            info += ["%.3f sec/step" % step_duration]
        if start_elapsed is not None:
            info += ["elapsed %s" % hms(start_elapsed)]
        if complete_frac not in (-1, None):
            assert 1 >= complete_frac > 0, f"{step} step, {complete_frac} complete_frac"
            assert start_elapsed is not None
            total_time_estimated = start_elapsed / complete_frac
            remaining_estimated = total_time_estimated - start_elapsed
            info += [
                "exp. remaining %s" % hms(remaining_estimated),
                "complete %.02f%%" % (complete_frac * 100),
            ]
        if start_elapsed is not None and complete_frac is None:
            info += ["(unk epoch len)"]
        print(", ".join(filter(None, info)), file=log.v5)


def _format_score(score: Dict[str, float]) -> str:
    """
    Like the TF engine format_score.

    :param score:
    :return: score(s) as str
    """
    if not score:
        return "None"
    if len(score) == 1:
        return _format_score_value(list(score.values())[0])
    return " ".join(["%s %s" % (k, _format_score_value(v)) for k, v in score.items()])


def _format_score_value(v: Any) -> str:
    if isinstance(v, float):
        if abs(v) > 1.0e3 or abs(v) < 1.0e-3:
            return f"{v:.3e}"
        else:
            return f"{v:.3f}"
    return str(v)


def _get_gpu_device() -> Optional[str]:
    if torch.cuda.is_available():
        return "cuda"
    if hasattr(torch.backends, "mps") and torch.backends.mps.is_available() and torch.backends.mps.is_built():
        return "mps"
    return None


def _get_batch_size_info(extern_data: TensorDict) -> Dict[str, int]:
    batch_dim = extern_data_util.get_batch_dim_from_extern_data(extern_data)
    info = {"num_seqs": int(batch_dim.get_dim_value())}
    covered_dims = {batch_dim}
    for k, v in extern_data.data.items():
        for dim in v.dims:
            if dim.is_dynamic() and dim not in covered_dims:
                covered_dims.add(dim)
                info[f"max_size:{dim.name}"] = int(dim.get_dim_value())
    return info


def get_device_from_config_opt(device: Optional[str]) -> ResultWithReason[str]:
    """
    :param device: as in config
    :return: resolved device
    """
    if os.environ.get("PT_DEVICE"):
        return ResultWithReason(os.environ["PT_DEVICE"], "PT_DEVICE env var")
    if not device:
        device = _get_gpu_device()
        if device:
            return ResultWithReason(device, "GPU automatically selected")
        return ResultWithReason("cpu", "no GPU found")
    reason = "config"
    if device == "gpu":
        device = _get_gpu_device()
        if not device:
            reasons = diagnose_gpu.diagnose_no_gpu()
            raise Exception("No GPU device found, but config requested 'gpu' device.\n" + "\n".join(reasons))
        reason = "'gpu' in config"
    return ResultWithReason(device, reason)


class _WrappedModuleRunStep(torch.nn.Module):
    """
    Wraps any Torch module (pure or RF),
    and the `forward` function calls the run step function (train_step or forward_step)
    and returns all produced raw tensors via the run context (losses or outputs) (:func:`rf.get_run_ctx`).
    This is useful to use the API of DistributedDataParallel and potentially other PyTorch modules.
    """

    def __init__(self, *, module: torch.nn.Module, engine: Engine):
        super().__init__()
        self.module = module
        self.engine = engine

    def forward(self, *args, **kwargs):
        """
        Call run step function (train_step or forward_step).

        :return: all produced raw tensors via the run context (:func:`rf.get_run_ctx`).
        """
        # noinspection PyProtectedMember
        self.engine._run_step(*args, **kwargs, _inside_wrapped=True)

        # Note that we don't directly use the returned raw values here,
        # but the PyTorch API might,
        # e.g. DistributedDataParallel checks it to collect all gradients.
        # We will use rf.get_run_ctx() later again in the engine to access these values.
        res = {}
        ctx = rf.get_run_ctx()
        for name, out in ctx.outputs.data.items():
            res["output/" + name] = out.raw_tensor
        for name, loss in ctx.losses.items():
            res["loss/" + name] = loss.loss.raw_tensor
        return res


@contextmanager
def _set_torch_default_dtype_ctx_mgr(dtype: torch.dtype):
    old_dtype = torch.get_default_dtype()
    try:
        torch.set_default_dtype(dtype)
        yield
    finally:
        torch.set_default_dtype(old_dtype)


def _parse_log_grad_norm(config: Config) -> Optional[Union[int, float]]:
    log_grad_norm = config.opt_typed_value("log_grad_norm", False)
    if isinstance(log_grad_norm, str):
        if log_grad_norm.lower() in ["true", "false", "none"]:
            log_grad_norm = {"true": True, "false": False, "none": None}[log_grad_norm.lower()]
        else:
            raise ValueError(f"Invalid value for log_grad_norm: {log_grad_norm!r}")
    if log_grad_norm is None:
        pass
    elif isinstance(log_grad_norm, bool):
        if log_grad_norm:
            log_grad_norm = 2
        else:
            log_grad_norm = None
    elif isinstance(log_grad_norm, (int, float)):
        assert log_grad_norm > 0, f"log_grad_norm {log_grad_norm} > 0 expected"  # otherwise fine...
    else:
        raise TypeError(f"Invalid type for log_grad_norm: {log_grad_norm!r} type {type(log_grad_norm)}")
    return log_grad_norm


def _get_total_grad_norm(model: torch.nn.Module, p: float) -> float:
    return float(
        torch.norm(
            torch.stack(
                [
                    param.grad.norm(p=p).detach().cpu()
                    for param in model.parameters()
                    if param.requires_grad and param.grad is not None
                ]
            ),
            p=p,
        ).item()
    )



"""
Test :mod:`returnn.torch.util`.
"""

from __future__ import annotations

import _setup_test_env  # noqa

import os
import sys
import unittest
import torch

from torch_utils import report_profile

from returnn.util import better_exchook


@unittest.skipIf(torch.__version__ < (2,), "gradient_checkpoint_scope needs PyTorch >= 2.0")
def test_gradient_checkpoint_scope():
    # https://github.com/rwth-i6/returnn/issues/1552
    from copy import deepcopy
    from torch.profiler import profile, record_function, ProfilerActivity
    from returnn.torch.util.gradient_checkpoint import gradient_checkpoint_scope

    shape = (101, 103)

    class _Model(torch.nn.Module):
        def __init__(self, *, use_grad_ckpt: bool = False):
            super().__init__()
            self.var = torch.nn.Parameter(torch.randn(shape))
            self.input_var = torch.nn.Parameter(torch.randn(shape))
            self.opt = torch.optim.SGD(self.parameters(), lr=0.1)  # not common to have this here but ok for the test
            self.use_grad_ckpt = use_grad_ckpt

        @staticmethod
        def get_var_noise() -> torch.Tensor:
            return torch.randn(shape)

        def forward(self, x: torch.Tensor) -> torch.Tensor:
            if not self.use_grad_ckpt:
                return (self.var + self.get_var_noise()) * x

            with gradient_checkpoint_scope():
                v_ = self.var + self.get_var_noise()
            return v_ * x

        def demo_run(self):
            x = self.input_var
            y = self(x)
            loss = y.sum()  # dummy loss
            del x, y  # not needed anymore. makes test cleaner.
            loss.backward()
            del loss  # not needed anymore
            self.opt.step()
            self.opt.zero_grad()

    model = _Model(use_grad_ckpt=False)
    param_state = deepcopy(model.state_dict())
    rng_state = torch.get_rng_state()
    with profile(activities=[ProfilerActivity.CPU], profile_memory=True, with_stack=True, record_shapes=True) as prof:
        with record_function("train_step_no_grad_ckpt"):
            model.demo_run()
    b = 4  # size single f32
    t = shape[0] * shape[1] * b  # size tensor
    r = rng_state.numel() * rng_state.element_size()
    report_profile(
        prof,
        [
            # ignore private calls
            # ignore torchop
            # ignore allocs/deallocs <=b
            ("pycall", {"callsite_name": "demo_run"}),
            ("pycall", {"callsite_name": "forward"}),
            ("pycall", {"callsite_name": "get_var_noise"}),
            ("alloc", {"name": "*/get_var_noise/aten::randn", "size": t, "total_alloc": t}),
            ("alloc", {"name": "*/forward/aten::add", "size": t, "total_alloc": t * 2}),
            ("dealloc", {"name": "*/get_var_noise/aten::randn", "size": -t, "total_alloc": t}),
            ("alloc", {"name": "*/forward/aten::mul", "size": t, "total_alloc": t * 2}),
            ("dealloc", {"name": "*/forward/aten::mul", "size": -t, "total_alloc": t}),
            ("pycall", {"callsite_name": "backward"}),
            ("alloc", {"name": "*/backward/*MulBackward0", "size": t, "total_alloc": t * 2}),
            ("alloc", {"name": "*/backward/*MulBackward0", "size": t, "total_alloc": t * 3}),
            ("dealloc", {"name": "*/forward/aten::add", "size": -t, "total_alloc": t * 2}),
            ("torchop", {"name": "Optimizer.step#SGD.step"}),
            ("torchop", {"name": "Optimizer.zero_grad#SGD.zero_grad"}),
            ("dealloc", {"name": "*/backward/*MulBackward0", "size": -t, "total_alloc": t}),
            ("dealloc", {"name": "*/backward/*MulBackward0", "size": -t, "total_alloc": 0}),
        ],
    )
    param_post_state = deepcopy(model.state_dict())

    print("**** now with grad chkpt ****")
    model = _Model(use_grad_ckpt=True)
    model.load_state_dict(param_state)
    torch.set_rng_state(rng_state)
    with profile(activities=[ProfilerActivity.CPU], profile_memory=True, with_stack=True, record_shapes=True) as prof:
        with record_function("train_step_grad_ckpt"):
            model.demo_run()
    report_profile(
        prof,
        [
            ("pycall", {"callsite_name": "demo_run"}),
            ("pycall", {"callsite_name": "forward"}),
            ("pycall", {"callsite_name": "get_var_noise"}),
            ("pycall", {"callsite_name": "__torch_dispatch__"}),
            ("alloc", {"name": "*/get_rng_state", "size": r, "total_alloc": r}),
            ("alloc", {"name": "*/get_var_noise/aten::randn", "size": t, "total_alloc": t + r}),
            ("pycall", {"callsite_name": "record_op"}),
            ("alloc", {"name": "*/forward/aten::add", "size": t, "total_alloc": t * 2 + r}),
            ("pycall", {"callsite_name": "record_op"}),
            ("dealloc", {"name": "*/get_var_noise/aten::randn", "size": -t, "total_alloc": t + r}),
            ("pycall", {"callsite_name": "_pack_hook"}),
            ("alloc", {"name": "*/forward/aten::mul", "size": t, "total_alloc": t * 2 + r}),
            # Not sure that we can always rely on this order here in the test...
            ("pycall", {"callsite_name": "_tensor_del_hook"}),
            ("pycall", {"callsite_name": "exit_saved_tensors_hooks_scope"}),
            ("pycall", {"callsite_name": "_custom_saved_tensors_hooks_exit"}),
            ("pycall", {"callsite_name": "_unregister_custom_saved_tensors_hooks"}),
            ("dealloc", {"name": "*/forward/aten::add", "size": -t, "total_alloc": t + r}),  # !!
            ("dealloc", {"name": "*/forward/aten::mul", "size": -t, "total_alloc": r}),
            ("pycall", {"callsite_name": "backward"}),
            ("pycall", {"callsite_name": "_unpack_hook"}),
            ("pycall", {"callsite_name": "maybe_recompute"}),
            ("pycall", {"callsite_name": "get_rng_state"}),
            ("alloc", {"name": "*/get_rng_state", "size": r, "total_alloc": r * 2}),
            ("pycall", {"callsite_name": "set_rng_state"}),
            ("pycall", {"callsite_name": "recompute"}),
            (
                "alloc",
                {
                    "name": "*/backward/_unpack_hook/maybe_recompute/recompute/aten::randn",
                    "size": t,
                    "total_alloc": t + r * 2,
                },
            ),
            ("alloc", {"name": "*/backward/*/recompute/aten::add", "size": t, "total_alloc": t * 2 + r * 2}),
            ("dealloc", {"name": "*/recompute/aten::randn", "size": -t, "total_alloc": t + r * 2}),
            ("pycall", {"callsite_name": "set_rng_state"}),
            ("dealloc", {"name": "*/backward/*/get_rng_state", "size": -r, "total_alloc": t + r}),
            ("dealloc", {"name": "*/forward/*/get_rng_state", "size": -r, "total_alloc": t}),
            ("alloc", {"name": "*/backward/*MulBackward0", "size": t, "total_alloc": t * 2}),
            ("alloc", {"name": "*/backward/*MulBackward0", "size": t, "total_alloc": t * 3}),
            ("dealloc", {"name": "*/recompute/aten::add", "size": -t, "total_alloc": t * 2}),
            ("torchop", {"name": "Optimizer.step#SGD.step"}),
            ("torchop", {"name": "Optimizer.zero_grad#SGD.zero_grad"}),
            ("dealloc", {"name": "*/backward/*MulBackward0", "size": -t, "total_alloc": t}),
            ("dealloc", {"name": "*/backward/*MulBackward0", "size": -t, "total_alloc": 0}),
        ],
    )
    param_post_state_ = deepcopy(model.state_dict())
    assert set(param_post_state.keys()) == set(param_post_state_.keys())
    for k in param_post_state.keys():
        torch.testing.assert_allclose(param_post_state[k], param_post_state_[k])


@unittest.skipIf(torch.__version__ < (2,), "gradient_checkpoint_scope needs PyTorch >= 2.0")
def test_gradient_checkpoint_scope_twice():
    # https://github.com/rwth-i6/returnn/issues/1579

    from returnn.torch.util.gradient_checkpoint import gradient_checkpoint_scope

    shape = (101, 103)

    class _Model(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.var = torch.nn.Parameter(torch.randn(shape))
            self.input_var = torch.nn.Parameter(torch.randn(shape))
            self.opt = torch.optim.SGD(self.parameters(), lr=0.1)  # not common to have this here but ok for the test

        def forward(self, x: torch.Tensor) -> torch.Tensor:
            return self.get_var() * x

        def get_var(self) -> torch.Tensor:
            with gradient_checkpoint_scope():
                return self.var + torch.randn(shape)

        def get_input(self) -> torch.Tensor:
            x = self.input_var
            with gradient_checkpoint_scope():
                return x + torch.randn(shape)

        def demo_run(self):
            self.opt.zero_grad()
            y = self(self.get_input())
            loss = y.sum()  # dummy loss
            del y  # not needed anymore
            loss.backward()
            del loss  # not needed anymore
            self.opt.step()

    orig_gradient_checkpoint_scope_tensor_del_hook = gradient_checkpoint_scope._tensor_del_hook
    try:
        # Overwrite this here to trigger the case where the tensor del hook will not do the cleanup.
        gradient_checkpoint_scope._tensor_del_hook = lambda self: None

        model = _Model()
        model.demo_run()
        model.demo_run()

    finally:
        gradient_checkpoint_scope._tensor_del_hook = orig_gradient_checkpoint_scope_tensor_del_hook


@unittest.skipIf(torch.__version__ < (2,), "gradient_checkpoint_scope needs PyTorch >= 2.0")
def test_saved_tensors_hooks_gc_segfault():
    # https://github.com/rwth-i6/returnn/issues/1581
    # https://github.com/pytorch/pytorch/issues/130734

    # noinspection PyProtectedMember
    from returnn.torch.util.gradient_checkpoint import _can_exit_saved_tensors_hooks_inside_hooks

    if not _can_exit_saved_tensors_hooks_inside_hooks():
        raise unittest.SkipTest("Not yet fixed.")

    shape = (101, 103)
    for i in range(10):
        print("**** iter", i)
        v = torch.nn.Parameter(torch.randn(shape))

        class _Handler:
            def __init__(self):
                self.scope = torch.autograd.graph.saved_tensors_hooks(self._pack_hook, self._unpack_hook)
                self.scope.__enter__()
                self.exited = False

            def _pack_hook(self, x):
                print(f"*** _pack_hook {self}")
                return x

            def _unpack_hook(self, x):
                print(f"*** _unpack_hook {self}")
                if not self.exited:
                    self.exited = True
                    print(
                        f"*** exit {self.scope},"
                        f" pack_hook {hex(id(self.scope.pack_hook))},"
                        f" unpack_hook {hex(id(self.scope.unpack_hook))}"
                    )
                    self.scope.__exit__()
                return x

        with torch.autograd.graph.saved_tensors_hooks(lambda x: x, lambda x: x):
            handler = _Handler()  # keep ref...  # noqa
            x = v * torch.randn(shape)
            x.sum().backward()


def test_debug_inf_nan():
    param = torch.nn.Parameter(torch.tensor([0.5349, 0.8094, -100, 0, -0.9890, 1, 1.3221, 0.8172, -0.7658, -0.7506]))

    def func():
        x = torch.tensor([0.5349, 0, 1.1103, -1.6898, -0.9890, 1, 1.3221, 0.8172, -0.7658, -0.7506])
        x = mod1(x)
        x = mod2(x)
        x = mod3(x) * param
        x = mod4(x)
        x = mod1(x)
        x = mod2(x)
        x = mod2(x)
        x = mod5(x)
        return x.sum()

    def mod1(x: torch.Tensor) -> torch.Tensor:
        return x * 2

    def mod2(x: torch.Tensor) -> torch.Tensor:
        return x.exp()

    def mod3(x: torch.Tensor) -> torch.Tensor:
        return x - 2

    def mod4(x: torch.Tensor) -> torch.Tensor:
        x.subtract_(-3.5)
        return x

    def mod5(x: torch.Tensor) -> torch.Tensor:
        return x / x

    x = func()
    print(x)
    print("inf/nan:", torch.isinf(x).any().item(), torch.isnan(x).any().item())

    from returnn.torch.util.debug_inf_nan import debug_inf_nan

    # Run directly, to just test that it goes through without exception.
    # For some reason, the detect_anomaly does not print the forward op?
    debug_inf_nan(func, with_grad=True, stop_reporting_after_first_inf_nan=False)

    from io import StringIO

    out = StringIO()
    debug_inf_nan(func, file=out, stop_reporting_after_first_inf_nan=False)
    assert "inf in aten.exp" in out.getvalue()
    assert "nan in aten.div" in out.getvalue()
    assert "mod5" in out.getvalue()
    assert os.path.basename(__file__) in out.getvalue()


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



import os
import numpy as np

from .bundle_file import BundleFile
from returnn.util.basic import long


class NormalizationData:
    """This class holds normalization data for inputs and outputs.
    It also contains methods to create the normalization HDF file.
    """

    GROUP_INPUTS = "inputs"
    GROUP_OUTPUTS = "outputs"

    DATASET_MEAN = "mean"
    DATASET_MEAN_OF_SQUARES = "meanOfSquares"
    DATASET_VARIANCE = "variance"
    DATASET_TOTAL_FRAMES = "totalNumberOfFrames"

    DATASET_TIME_DIMENSION_INDEX = 0
    DATASET_FEATURE_DIMENSION_INDEX = 1

    SUMMATION_PRECISION = 1e-5

    @staticmethod
    def createNormalizationFile(bundleFilePath, outputFilePath, dtype=np.float64, flag_includeOutputs=True):
        """Calculates means over inputs and outputs of datasets in the HDF files
        described by the given bundle file.

        :see: BundleFile.BundleFile

        Each HDF dataset file is expected to have the following groups:

          * NormalizationData.GROUP_INPUTS (the group for the input data)
          * NormalizationData.GROUP_OUTPUTS (the group for the output data)

        Each group may have datasets. Each dataset is expected to have
        shape (time frames, features).
        E.g. (267, 513) -- 267 time frames each containing a feature vector of
        dimensionality 513.

        The method writes results into the given output file.
        Availability of means and variances depends on whether the corresponding
        groups are available in the input dataset HDF files.

        !!! IMPORTANT !!!
        General rule of thumb: if one dataset file has both input and output
        groups then you should make sure that all the dataset files have them.
        Otherwise means and variance will not be correct.
        It is OK if *all* the datasets have only the input group.
        In this case means and variance only for inputs will be calculated.

        :type bundleFilePath: str
        :param bundleFilePath: path to the bundle file. :see: BundleFile.BundleFile
        :type outputFilePath: str
        :param outputFilePath: path to the output HDF normalization file.
        :type dtype: numpy.dtype
        :param dtype: type of data to use during calculations.
        :type flag_includeOutputs: bool
        :param flag_includeOutputs: if True then normalization data will be
                                    calculated for outputs (targets) as well.
        """
        NormalizationData._calculateNormalizationData(
            bundleFilePath, outputFilePath, NormalizationData.GROUP_INPUTS, dtype=dtype
        )
        if flag_includeOutputs:
            NormalizationData._calculateNormalizationData(
                bundleFilePath, outputFilePath, NormalizationData.GROUP_OUTPUTS, dtype=dtype
            )

    @staticmethod
    def _calculateNormalizationData(bundleFilePath, outputFilePath, groupName, dtype=np.float64):
        """Helper method.
        Calculates and writes into the output HDF file mean, mean of squares,
        variance and total number of frames for the datasets in the given HDF
        group.

        :type bundleFilePath: str
        :param bundleFilePath: path to the bundle file. :see: BundleFile.BundleFile
        :type outputFilePath: str
        :param outputFilePath: path to the output HDF normalization file. If file
                               already exists it will not be truncated.
        :type groupName: str
        :param groupName: name of the HDF group for which normalization data
                          should be calculated. Also, a group with this name will
                          be created in the output HDF file to store the calculated
                          normalization data.
        :type dtype: numpy.dtype
        :param dtype: type of data to use during calculations.
        """
        import h5py

        accumulatedSum = None
        accumulatedSumOfSqr = None
        totalFrames = long()
        bundle = BundleFile(bundleFilePath)
        for filePath in bundle.datasetFilePaths:
            with h5py.File(filePath, mode="r") as datasetFile:
                intermSum, intermSumOfSqr, intermTotalFrames = NormalizationData._accumulateSums(
                    datasetFile, groupName, dtype=dtype
                )
                accumulatedSum = NormalizationData._updateTotalSum(accumulatedSum, intermSum)
                accumulatedSumOfSqr = NormalizationData._updateTotalSum(accumulatedSumOfSqr, intermSumOfSqr)
                totalFrames += intermTotalFrames

        mean, meanOfSquares, variance = NormalizationData._calculateMeans(
            accumulatedSum, accumulatedSumOfSqr, totalFrames
        )

        with h5py.File(outputFilePath, mode="a") as out:
            NormalizationData._writeData(out, groupName, mean, meanOfSquares, variance, totalFrames, dtype=dtype)

    @staticmethod
    def _accumulateSums(f, groupName, dtype=np.float64):
        """Helper method.
        Accumulate sums and sums of squares over feature vectors for a given group.

        :type f: h5py.File
        :param f: handle to an opened HDF file with datasets
        :type groupName: str
        :param groupName: HDF group containing datasets
        :type dtype: numpy.dtype
        :param dtype: type of data to use during calculations.
        :rtype: tuple (numpy.ndarray | None, numpy.ndarray | None, long)
        :return: tuple (sum, sum of squares, total number of time frames)
                 if they are available
        """
        sum = None
        sumOfSqr = None
        totalFrames = np.int64(0)
        if groupName not in f:
            return sum, sumOfSqr, totalFrames
        group = f[groupName]
        datasetNames = group.keys()
        if len(datasetNames) == 0:
            return sum, sumOfSqr, totalFrames
        featDims = group[datasetNames[0]].shape[NormalizationData.DATASET_FEATURE_DIMENSION_INDEX]
        sum = np.zeros(featDims, dtype=dtype)
        sumOfSqr = np.zeros(featDims, dtype=dtype)
        for dsName in datasetNames:
            dataset = group[dsName][...]
            sum += np.sum(dataset, axis=NormalizationData.DATASET_TIME_DIMENSION_INDEX)
            sumOfSqr += np.sum(np.square(dataset), axis=NormalizationData.DATASET_TIME_DIMENSION_INDEX)
            totalFrames += dataset.shape[NormalizationData.DATASET_TIME_DIMENSION_INDEX]
        return sum, sumOfSqr, totalFrames

    @staticmethod
    def _updateTotalSum(totalSum, intermediateSum):
        """Helper method.
        Updates total sum with intermediate sum if the latter is available.

        :type totalSum: numpy.ndarray | None
        :param totalSum: total sum
        :type intermediateSum: numpy.ndarray | None
        :param intermediateSum: intermediate sum
        :rtype: numpy.ndarray | None
        :return: updated total sum if available
        """
        if totalSum is None and intermediateSum is None:
            return None
        if totalSum is None:
            return intermediateSum
        if intermediateSum is None:
            return totalSum
        # floating point summation check
        oldSum = totalSum
        newSum = np.add(totalSum, intermediateSum)
        sumErr = np.sum(np.abs(newSum - oldSum - intermediateSum))
        if sumErr > NormalizationData.SUMMATION_PRECISION:
            raise FloatingPointError(
                "sums have very different orders of magnitude." " summation error = {}".format(sumErr)
            )
        return newSum

    @staticmethod
    def _calculateMeans(totalSum, totalSumOfSqr, totalFrames):
        """Helper method.
        Calculate mean, mean of squares and variance if they are available.

        :type totalSum: numpy.ndarray | None
        :param totalSum: total sum of features
        :type totalSumOfSqr: numpy.ndarray | None
        :param totalSumOfSqr: total sum of squares of features
        :type totalFrames: long
        :param totalFrames: total number of timeframes
        :rtype: tuple (numpy.ndarray | None, numpy.ndarray | None, numpy.ndarray | None)
        :return: tuple (mean, mean of squares, variance) if they are available
        """
        mean = None
        meanOfSquares = None
        variance = None
        if totalSum is not None:
            assert totalFrames > 0
            mean = totalSum / totalFrames
        if mean is not None and totalSumOfSqr is not None:
            assert totalFrames > 0
            meanOfSquares = totalSumOfSqr / totalFrames
            # Var[X] = E[X ^ 2] - (E[X]) ^ 2
            variance = meanOfSquares - np.square(mean)
        return mean, meanOfSquares, variance

    @staticmethod
    def _writeData(f, groupName, mean, meanOfSqr, variance, totalFrames, dtype=np.float64):
        """Helper method.
        Writes means and variance for a given group.

        :type f: h5py.File
        :param f: handle to an opened HDF file to which data should be written.
        :type groupName: str
        :param groupName: HDF group name
        :type mean: numpy.ndarray | None
        :param mean: mean
        :type meanOfSqr: numpy.ndarray | None
        :param meanOfSqr: mean of squares
        :type variance: numpy.ndarray | None
        :param variance: variance
        :type totalFrames: long
        :param totalFrames: total number of time frames
        :type dtype: numpy.dtype
        :param dtype: type of data to use for writing the data
        """
        if groupName in f:
            del f[groupName]
        group = f.create_group(groupName)
        dsNames = [
            NormalizationData.DATASET_MEAN,
            NormalizationData.DATASET_MEAN_OF_SQUARES,
            NormalizationData.DATASET_VARIANCE,
        ]
        datasets = [mean, meanOfSqr, variance]
        for name, ds in zip(dsNames, datasets):
            NormalizationData._writeDataset(group, name, ds, dtype)
        if totalFrames > 0:
            group.create_dataset(NormalizationData.DATASET_TOTAL_FRAMES, data=totalFrames)

    @staticmethod
    def _writeDataset(group, datasetName, dataset, dtype=np.float64):
        """Helper Method.
        Writes dataset into an HDF group if the dataset is available.

        :type group: h5py.Group
        :param group: HDF group handle
        :type datasetName: str
        :param datasetName: name of the dataset
        :type dataset: numpy.ndarray | None
        :param dataset: actual data of the dataset
        :type dtype: numpy.dtype
        :param dtype: type of data to use for writing the data.
        """
        if dataset is None:
            return
        group.create_dataset(datasetName, data=dataset, dtype=dtype)

    def __init__(self, normalizationFilePath):
        """Reads normalization data from the given HDF file and saves it
        into the member variables.

        :type normalizationFilePath: str
        :param normalizationFilePath: path to the HDF file with normalization data.
        """
        self._normalizationFilePath = normalizationFilePath
        self._inputMean = None
        self._inputVariance = None
        self._outputMean = None
        self._outputVariance = None
        self._readNormalizationData()

    def _readNormalizationData(self):
        """Reads normalization data from the given HDF file.
        The file is expected to have the following structure.

        It may have two groups:
          * NormalizationData.GROUP_INPUTS (the group for the input data)
          * NormalizationData.GROUP_OUTPUTS (the group for the output data)

        Each group may have two datasets:
          * NormalizationData.DATASET_MEAN (the dataset for mean)
          * NormalizationData.DATASET_VARIANCE (the dataset for variance)

        Everything is optional e.g. when only the group for the input data
        is present and it contains only the dataset for mean then only this
        data will be read. No exception will be thrown.

        The groups may also contain additional optional information such as
        e.g. total number of time frames, mean of squares etc.
        However, this information is not read here.
        """
        import h5py

        if not os.path.isfile(self._normalizationFilePath):
            raise IOError(self._normalizationFilePath + " does not exist")
        with h5py.File(self._normalizationFilePath, mode="r") as f:
            self._inputMean, self._inputVariance = self._getMeanAndVarianceFromGroup(f, self.GROUP_INPUTS)
            self._outputMean, self._outputVariance = self._getMeanAndVarianceFromGroup(f, self.GROUP_OUTPUTS)

    @staticmethod
    def _getMeanAndVarianceFromGroup(f, groupName):
        """Reads mean and variance from the given group if they are available.
        Both mean and variance are optional i.e. they may be absent in the
        given HDF group.

        :type f: h5py.File
        :param f: handle to an opened HDF file with normalization data.
        :type groupName: str
        :param groupName: name of the HDF group from which mean and variance
                          should be read.
        :rtype: tuple (numpy.ndarray | None, numpy.ndarray | None)
        :return: a tuple (mean, variance) each of which may be None
                 if the data is not available.
        """
        mean = None
        variance = None
        if groupName not in f:
            return mean, variance
        group = f[groupName]
        if NormalizationData.DATASET_MEAN in group:
            mean = group[NormalizationData.DATASET_MEAN][...]
        if NormalizationData.DATASET_VARIANCE in group:
            variance = group[NormalizationData.DATASET_VARIANCE][...]
        return mean, variance

    @property
    def inputMean(self):
        """Mean of the input data.

        :rtype: numpy.ndarray | None
        :return: Mean of the input data if it is available or None otherwise.
        """
        return self._inputMean

    @property
    def inputVariance(self):
        """Variance of the input data.

        :rtype: numpy.ndarray | None
        :return: Variance of the input data if it is available or None otherwise.
        """
        return self._inputVariance

    @property
    def outputMean(self):
        """Mean of the output data.

        :rtype: numpy.ndarray | None
        :return: Mean of the output data if it is available or None otherwise.
        """
        return self._outputMean

    @property
    def outputVariance(self):
        """Variance of the output data.

        :rtype: numpy.ndarray | None
        :return: Variance of the output data if it is available or None otherwise.
        """
        return self._outputVariance



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author: Pavel Golik (golik@cs.rwth-aachen.de)

"""
This module is about reading (maybe later also writing) the Sprint archive format.
"""

from __future__ import annotations
from typing import List
import sys
import os
import typing
import array
from struct import pack, unpack
import numpy
import zlib
import mmap


class FileInfo:
    """
    File info.
    """

    def __init__(self, name, pos, size, compressed, index):
        """
        :param str name:
        :param int pos:
        :param int size:
        :param bool|int compressed:
        :param int index:
        """
        self.name = name
        self.pos = pos
        self.size = size
        self.compressed = compressed
        self.index = index

    def __repr__(self):
        return "FileInfo(%s)" % " ".join(str(s) for s in self.__dict__.values())


class FileArchive:
    """
    File archive.
    """

    # read routines
    def read_u32(self):
        """
        :rtype: int
        """
        return int(unpack("i", self.f.read(4))[0])

    # noinspection PyPep8Naming
    def read_U32(self):
        """
        :rtype: int
        """
        return int(unpack("I", self.f.read(4))[0])

    # noinspection PyPep8Naming
    def read_U8(self):
        """
        :rtype: int
        """
        return int(unpack("B", self.f.read(1))[0])

    # noinspection PyPep8Naming
    def read_packed_U32(self):
        """
        :rtype: int
        """
        result = 0
        s = 0
        while True:
            b = self.read_U8()
            if (b & 0b10000000) == 0:
                result |= b << s
            else:
                result |= (b & 0b01111111) << s
            s += 7
            if (b & 0b10000000) == 0:
                break
        return result

    def read_u64(self):
        """
        :rtype: int
        """
        return int(unpack("q", self.f.read(8))[0])

    def read_char(self):
        """
        :rtype: int
        """
        return unpack("b", self.f.read(1))[0]

    def read_bytes(self, length):
        """
        :param int length:
        :rtype: bytes
        """
        return unpack("%ds" % length, self.f.read(length))[0]

    def read_str(self, length, encoding="ascii"):
        """
        :param int length:
        :param str encoding:
        :rtype: str
        """
        return self.read_bytes(length).decode(encoding)

    # noinspection PyPep8Naming
    def read_S16(self):
        """
        :rtype: float
        """
        return float(unpack("H", self.f.read(2))[0])

    def read_f32(self):
        """
        :rtype: float
        """
        return float(unpack("f", self.f.read(4))[0])

    def read_f64(self):
        """
        :rtype: float
        """
        return float(unpack("d", self.f.read(8))[0])

    def read_v(self, typ, size):
        """
        :param str typ: "f" for float (float32) or "d" for double (float64)
        :param int size: number of elements to return
        :return: numpy array of shape (size,) of dtype depending on typ
        :rtype: numpy.ndarray
        """
        if typ == "f":
            b = 4
            t = numpy.float32
        elif typ == "d":
            b = 8
            t = numpy.float64
        else:
            raise NotImplementedError("typ: %r" % typ)
        if isinstance(self.f, mmap.mmap):
            res = numpy.frombuffer(self.f, t, size, self.f.tell())
            self.f.seek(b * size, os.SEEK_CUR)
        else:
            res = numpy.fromfile(self.f, t, size, "")
        return res

    # write routines
    def write_str(self, s, enc="ascii"):
        """
        :param str s:
        :param str enc:
        :rtype: int
        """
        return self.f.write(pack("%ds" % len(s.encode(enc)), s.encode(enc)))

    def write_char(self, i):
        """
        :param int i:
        :rtype: int
        """
        return self.f.write(pack("b", i))

    def write_u32(self, i):
        """
        :param int i:
        :rtype: int
        """
        return self.f.write(pack("i", i))

    # noinspection PyPep8Naming
    def write_U32(self, i):
        """
        :param int i:
        :rtype: int
        """
        return self.f.write(pack("I", i))

    def write_u64(self, i):
        """
        :param int i:
        :rtype: int
        """
        return self.f.write(pack("q", i))

    def write_f32(self, i):
        """
        :param float i:
        :rtype: int
        """
        return self.f.write(pack("f", i))

    def write_f64(self, i):
        """
        :param float i:
        :rtype: int
        """
        return self.f.write(pack("d", i))

    SprintCacheHeader = "SP_ARC1\0"
    start_recovery_tag = 0xAA55AA55
    end_recovery_tag = 0x55AA55AA

    def __init__(self, filename, must_exists=True, encoding="ascii"):
        self.encoding = encoding

        self.ft = {}  # type: typing.Dict[str,FileInfo]
        if os.path.exists(filename):
            self.allophones = []
            self.f = open(filename, "rb")
            header = self.read_str(len(self.SprintCacheHeader))
            assert header == self.SprintCacheHeader

            ft = bool(self.read_char())
            if ft:
                self.read_file_info_table()
            else:
                self.scan_archive()

        else:
            assert not must_exists, "File does not exist: %r" % filename
            self.f = open(filename, "wb")
            self.write_str(self.SprintCacheHeader)
            self.write_char(1)

        self._short_seg_names = {os.path.basename(n): n for n in self.ft.keys()}
        if len(self._short_seg_names) < len(self.ft):
            # We don't have a unique mapping, so we cannot use this.
            self._short_seg_names.clear()

    def file_list(self):
        """
        :rtype: list[str]
        """
        return self.ft.keys()

    def finalize(self):
        """
        Finalize.
        """
        self.write_file_info_table()

    def read_file_info_table(self):
        """
        Read file info table.
        """
        self.f.seek(-8, 2)
        pos_count = self.read_u64()
        self.f.seek(pos_count)
        count = self.read_u32()
        if not count > 0:
            return
        for i in range(count):
            str_len = self.read_u32()
            name = self.read_str(str_len, self.encoding)
            pos = self.read_u64()
            size = self.read_u32()
            comp = self.read_u32()
            self.ft[name] = FileInfo(name, pos, size, comp, i)
            # TODO: read empty files

    def write_file_info_table(self):
        """
        Write file info table.
        """
        pos = self.f.tell()
        self.write_u32(len(self.ft))

        for fi in self.ft.values():
            self.write_u32(len(fi.name.encode(self.encoding)))
            self.write_str(fi.name, self.encoding)
            self.write_u64(fi.pos)
            self.write_u32(fi.size)
            self.write_u32(fi.compressed)

        self.write_u64(0)
        self.write_u64(pos)

    def scan_archive(self):
        """
        Scan archive.
        """
        i = 0
        self.f.seek(0, 2)
        end = self.f.tell()
        self.f.seek(0)
        while self.f.tell() < end:
            if self.read_U32() != self.start_recovery_tag:
                continue

            fn_len = self.read_u32()
            name = self.read_str(fn_len, self.encoding)
            pos = self.f.tell()
            size = self.read_u32()
            comp = self.read_u32()
            self.read_u32()  # chk

            self.f.seek(size, 1)
            self.ft[name] = FileInfo(name, pos, size, comp, i)
            i += 1

            self.read_U32()  # end tag

            # raise NotImplementedError("Need to scan archive if no "
            #                           "file info table found.")

    def _raw_read(self, size, typ):
        """
        :param int|None size: needed for typ == "str"
        :param str typ: "str", "feat" or "align"
        :return: depending on typ, "str" -> string, "feat" -> (time, data), "align" -> align,
          where string is a str,
          time is list of time-stamp tuples (start-time,end-time) in millisecs,
            data is a list of features, each a numpy vector,
          align is a list of (time, allophone, state, weight), time is an int from 0 to len of align,
            allophone is some int, state is e.g. in [0,1,2].
        :rtype: str|(list[numpy.ndarray],list[numpy.ndarray])|list[(int,int,int,float)]
        """

        if typ == "str":
            return self.read_str(size, self.encoding)

        elif typ == "feat":
            type_len = self.read_U32()
            type_ = self.read_str(type_len)
            # print(typ)
            assert type_ == "vector-f32"
            count = self.read_U32()
            data = [None] * count  # type: typing.List[typing.Optional[numpy.ndarray]]
            time_ = [None] * count  # type: typing.List[typing.Optional[numpy.ndarray]]
            for i in range(count):
                size = self.read_U32()
                data[i] = self.read_v("f", size)  # size x f32
                time_[i] = self.read_v("d", 2)  # 2    x f64
            return time_, data

        elif typ in ["align", "align_raw"]:
            type_len = self.read_U32()
            type_ = self.read_str(type_len)
            assert type_ == "flow-alignment"
            self.read_u32()  # flag ?
            alignment_header = self.read_str(8)
            if alignment_header in ["ALIGNRLE", "AALPHRLE"]:
                # In case of AALPHRLE, after the alignment, we include the alphabet of the used labels.
                # We ignore this at the moment.
                size = self.read_U32()
                if size < (1 << 31):
                    # RLE scheme
                    time = 0
                    alignment = []
                    while len(alignment) < size:
                        n = self.read_char()
                        # print(n)
                        if n > 0:
                            while n > 0:
                                mix, state = self.read_u32(), None
                                if typ != "align_raw":
                                    mix, state = self.get_state(mix)
                                # print(mix, state)
                                # print(time, self.allophones[mix])
                                alignment.append((time, mix, state, 1))
                                time += 1
                                n -= 1
                        elif n < 0:
                            mix, state = self.read_u32(), None
                            if typ != "align_raw":
                                mix, state = self.get_state(mix)
                            while n < 0:
                                # print(mix, state)
                                # print(time, self.allophones[mix])
                                alignment.append((time, mix, state, 1))
                                time += 1
                                n += 1
                        else:
                            time = self.read_u32()
                            # print("time", time)
                    return alignment
                else:
                    # weighted RLE scheme
                    version = size
                    version &= (1 << 31) - 1
                    assert version <= 2
                    size = self.read_packed_U32()
                    t = 0
                    alignment = []
                    while len(alignment) < size:
                        num_items = self.read_packed_U32()
                        if num_items & 1:
                            t = self.read_packed_U32()
                        num_items /= 2
                        while num_items > 0:
                            emission = self.read_packed_U32()
                            mix, state = self.get_state(emission)

                            if version == 1:
                                weight = self.read_S16() / 65535  # not tested yet
                            elif version == 2:
                                weight = self.read_f32()
                            else:
                                raise NotImplementedError(f"Version {version} not implemented.")
                            num_items -= 1
                            alignment.append((t, mix, state, weight))
                        t += 1
                    return alignment
            else:
                raise Exception("No valid alignment header found (found: %r). Wrong cache?" % alignment_header)

    def has_entry(self, filename):
        """
        :param str filename: argument for self.read()
        :return: True if we have this entry
        """
        return filename in self.ft

    def read(self, filename, typ):
        """
        :param str filename: the entry-name in the archive
        :param str typ: "str", "feat" or "align"
        :return: depending on typ, "str" -> string, "feat" -> (time, data), "align" -> align,
          where string is a str,
          time is list of time-stamp tuples (start-time,end-time) in millisecs,
            data is a list of features, each a numpy vector,
          align is a list of (time, allophone, state, weight), time is an int from 0 to len of align,
            allophone is some int, state is e.g. in [0,1,2].
        :rtype: str|(list[numpy.ndarray],list[numpy.ndarray])|list[(int,int,int,float)]
        """

        if filename not in self.ft:
            if filename in self._short_seg_names:
                filename = self._short_seg_names[filename]

        fi = self.ft[filename]
        self.f.seek(fi.pos)
        size = self.read_U32()
        comp = self.read_U32()
        self.read_U32()  # chk
        if size == 0:
            return None

        if comp > 0:
            # read compressed bytes into memory as 'bytearray'
            a = array.array("b")
            a.fromfile(self.f, comp)
            # unpack
            b = zlib.decompress(a.tostring(), 15 + 32)
            # substitute self.f by an anonymous memmap file object
            # restore original file handle after we're done
            backup_f = self.f
            self.f = mmap.mmap(-1, len(b))
            self.f.write(b)
            self.f.seek(0)
            try:
                return self._raw_read(size=fi.size, typ=typ)
            finally:
                self.f = backup_f

        return self._raw_read(size=fi.size, typ=typ)

    def get_state(self, mix):
        """
        :param int mix:
        :return: (mix, state)
        :rtype: (int,int)
        """
        # See src/Tools/Archiver/Archiver.cc:getStateInfo() from Sprint source code.
        assert self.allophones
        max_states = 6
        state = None
        for state in range(max_states):
            if mix >= len(self.allophones):
                mix -= 1 << 26
            else:
                break
        assert mix >= 0
        return mix, state

    def set_allophones(self, f):
        """
        :param str f: allophone filename. line-separated. will ignore lines starting with "#"
        """
        del self.allophones[:]
        for line in open(f):
            line = line.strip()
            if line.startswith("#"):
                continue
            self.allophones.append(line)

    def get_allophones_list(self) -> List[str]:
        """
        :return: list of allophones
        """
        return self.allophones

    def add_feature_cache(self, filename, features, times):
        """
        :param str filename:
        :param features:
        :param times:
        """
        self.write_U32(self.start_recovery_tag)
        self.write_u32(len(filename.encode(self.encoding)))
        self.write_str(filename, self.encoding)
        pos = self.f.tell()
        if len(features) > 0:
            dim = len(features[0])
            size = 4 + 10 + 4 + len(features) * (4 + dim * 4 + 2 * 8)
            duration = times[-1][1]
        else:
            dim = 0
            size = 4 + 10 + 4
            duration = 1
        self.write_u32(size)
        self.write_u32(0)
        self.write_u32(0)

        self.write_u32(10)
        self.write_str("vector-f32")
        assert len(features) == len(times)
        self.write_u32(len(features))
        for f, t in zip(features, times):
            self.write_u32(len(f))
            for x in f.flat:
                self.write_f32(x)
            self.write_f64(t[0])
            self.write_f64(t[1])

        self.ft[filename] = FileInfo(filename, pos, size, 0, len(self.ft))
        self.write_U32(self.end_recovery_tag)

        self.add_attributes(filename, dim, duration)

    def add_attributes(self, filename, dim, duration):
        """
        :param str filename:
        :param int dim:
        :param float duration:
        """
        data = (
            "<flow-attributes>"
            '<flow-attribute name="datatype" value="vector-f32"/>'
            '<flow-attribute name="sample-size" value="%d"/>'
            '<flow-attribute name="total-duration" value="%.5f"/>'
            "</flow-attributes>"
        ) % (dim, duration)
        self.write_U32(self.start_recovery_tag)
        filename = "%s.attribs" % filename
        self.write_u32(len(filename.encode(self.encoding)))
        self.write_str(filename, self.encoding)
        pos = self.f.tell()
        size = len(data)
        self.write_u32(size)
        self.write_u32(0)
        self.write_u32(0)
        self.write_str(data)
        self.write_U32(self.end_recovery_tag)
        self.ft[filename] = FileInfo(filename, pos, size, 0, len(self.ft))


class FileArchiveBundle:
    """
    File archive bundle.
    """

    def __init__(self, filename=None, encoding="ascii"):
        """
        :param str|None filename: .bundle file
        :param str encoding: encoding used in the files
        """
        # filename -> FileArchive
        self.archives = {}  # type: typing.Dict[str,FileArchive]
        # archive content file -> FileArchive
        self.files = {}  # type: typing.Dict[str,FileArchive]
        self._short_seg_names = {}
        if filename is not None:
            self.add_bundle(filename=filename, encoding=encoding)

    def add_bundle(self, filename, encoding="ascii"):
        """
        :param str filename: bundle
        :param str encoding:
        """
        file_dir = os.path.dirname(filename) or "."
        for line in open(filename).read().splitlines():
            if line.startswith("/"):
                archive_file = line
            else:
                archive_file = f"{file_dir}/{line}"
            self.add_archive(filename=archive_file, encoding=encoding)

    def add_archive(self, filename, encoding="ascii"):
        """
        :param str filename: single archive
        :param str encoding:
        """
        if filename in self.archives:
            return
        self.archives[filename] = a = FileArchive(filename, must_exists=True, encoding=encoding)
        for f in a.ft.keys():
            self.files[f] = a
        # noinspection PyProtectedMember
        self._short_seg_names.update(a._short_seg_names)

    def add_bundle_or_archive(self, filename, encoding="ascii"):
        """
        :param str filename:
        :param str encoding:
        """
        if filename.endswith(".bundle"):
            self.add_bundle(filename, encoding=encoding)
        else:
            self.add_archive(filename, encoding=encoding)

    def file_list(self):
        """
        :rtype: list[str]
        :returns: list of content-filenames (which can be used for self.read())
        """
        return self.files.keys()

    def has_entry(self, filename):
        """
        :param str filename: argument for self.read()
        :return: True if we have this entry
        """
        return filename in self.files

    def read(self, filename, typ):
        """
        :param str filename: the entry-name in the archive
        :param str typ: "str", "feat" or "align"
        :return: depending on typ, "str" -> string, "feat" -> (time, data), "align" -> align,
          where string is a str,
          time is list of time-stamp tuples (start-time,end-time) in millisecs,
            data is a list of features, each a numpy vector,
          align is a list of (time, allophone, state, weight), time is an int from 0 to len of align,
            allophone is some int, state is e.g. in [0,1,2].
        :rtype: str|(list[numpy.ndarray],list[numpy.ndarray])|list[(int,int,int,float)]

        Uses FileArchive.read().
        """
        if filename not in self.files:
            if filename in self._short_seg_names:
                filename = self._short_seg_names[filename]
        return self.files[filename].read(filename, typ)

    def set_allophones(self, filename):
        """
        :param str filename: allophone filename
        """
        for a in self.archives.values():
            a.set_allophones(filename)

    def get_allophones_list(self) -> List[str]:
        """
        :return: list of allophones
        """
        a = next(iter(self.archives.values()))
        return a.get_allophones_list()


def open_file_archive(archive_filename, must_exists=True, encoding="ascii"):
    """
    :param str archive_filename:
    :param bool must_exists:
    :param str encoding:
    :rtype: FileArchiveBundle|FileArchive
    """
    if archive_filename.endswith(".bundle"):
        assert must_exists
        return FileArchiveBundle(archive_filename, encoding=encoding)
    else:
        return FileArchive(archive_filename, must_exists=must_exists, encoding=encoding)


def is_sprint_cache_file(filename):
    """
    :param str filename: file to check. must exist
    :return: True iff this is a sprint cache (which can be loaded with `open_file_archive()`)
    :rtype: bool
    """
    assert os.path.exists(filename)
    if not os.path.isfile(filename):
        return False
    if filename.endswith(".bundle"):
        return True
    with open(filename, "rb") as f:
        header_len = len(FileArchive.SprintCacheHeader)
        bs = unpack("%ds" % header_len, f.read(header_len))[0]
        header = bs.decode("ascii")
        return header == FileArchive.SprintCacheHeader


class AllophoneLabeling:
    """
    Allophone labeling.
    """

    def __init__(self, silence_phone, allophone_file, phoneme_file=None, state_tying_file=None, verbose_out=None):
        """
        :param str silence_phone: e.g. "si"
        :param str allophone_file: list of allophones
        :param str|None phoneme_file: list of phonemes
        :param str|None state_tying_file: allophone state tying (e.g. via CART).
            maps each allophone state to a class label
        :param file verbose_out: stream to dump log messages
        """
        assert phoneme_file or state_tying_file
        self.allophone_file = allophone_file
        self.allophones = [line for line in open(allophone_file).read().splitlines() if line and line[0] != "#"]
        self.allophones_idx = {p: i for i, p in enumerate(self.allophones)}
        self.sil_allo_state_id = self.allophones_idx[silence_phone + "{#+#}@i@f"]
        if verbose_out:
            print("AllophoneLabeling: Num allophones: %i" % len(self.allophones), file=verbose_out)
        self.sil_label_idx = None
        self.num_labels = None
        self.phonemes = None
        self.phoneme_idxs = None
        self.state_tying = None
        self.state_tying_by_allo_state_idx = None
        self.num_allo_states = None
        if phoneme_file:
            self.phonemes = open(phoneme_file).read().splitlines()
            self.phoneme_idxs = {p: i for i, p in enumerate(self.phonemes)}
            if not state_tying_file:
                self.sil_label_idx = self.phoneme_idxs[silence_phone]
                self.num_labels = len(self.phonemes)
                if verbose_out:
                    print("AllophoneLabeling: %i phones = labels." % self.num_labels, file=verbose_out)
        if state_tying_file:
            self.state_tying = {
                k: int(v) for line in open(state_tying_file).read().splitlines() for (k, v) in [line.split()]
            }
            self.sil_label_idx = self.state_tying[silence_phone + "{#+#}@i@f.0"]
            self.num_allo_states = self._get_num_allo_states()
            self.state_tying_by_allo_state_idx = {
                a + s * (1 << 26): self.state_tying["%s.%i" % (a_s, s)]
                for (a, a_s) in enumerate(self.allophones)
                for s in range(self.num_allo_states)
                if ("%s.%i" % (a_s, s)) in self.state_tying
            }
            self.num_labels = max(self.state_tying.values()) + 1
            if verbose_out:
                print("AllophoneLabeling: State tying with %i labels." % self.num_labels, file=verbose_out)
        assert self.num_labels is not None
        assert self.state_tying or self.phoneme_idxs

    def _get_num_allo_states(self):
        assert self.state_tying
        return max([int(s.split(".")[-1]) for s in self.state_tying.keys()]) + 1

    def get_label_idx_by_allo_state_idx(self, allo_state_idx):
        """
        :param int allo_state_idx:
        :rtype: int
        """
        if self.state_tying_by_allo_state_idx:
            return self.state_tying_by_allo_state_idx[allo_state_idx]
        # See getState above().
        max_states = 6
        allo_idx = allo_state_idx
        state_idx = 0
        for state_idx in range(max_states):
            if allo_idx >= len(self.allophones):
                allo_idx -= 1 << 26
            else:
                break
        assert allo_idx >= 0
        return self.get_label_idx(allo_idx, state_idx)

    def get_label_idx(self, allo_idx, state_idx):
        """
        :param int allo_idx:
        :param int state_idx:
        :rtype: int
        """
        if self.state_tying_by_allo_state_idx:
            try:
                return self.state_tying_by_allo_state_idx[allo_idx + state_idx * (1 << 26)]
            except KeyError:
                allo_str = self.allophones[allo_idx]
                r = self.state_tying.get("%s.%i" % (allo_str, state_idx))
                raise KeyError(
                    "allo idx %i (%r), state idx %i not found; entry: %r" % (allo_idx, allo_str, state_idx, r)
                )
        allo_str = self.allophones[allo_idx]
        phone = allo_str[: allo_str.index("{")]
        return self.phoneme_idxs[phone]


###############################################################################


class MixtureSet:
    """
    Mixture set.
    """

    # read routines
    def read_u32(self):
        """
        :rtype: int
        """
        return int(unpack("i", self.f.read(4))[0])

    # noinspection PyPep8Naming
    def read_U32(self):
        """
        :rtype: int
        """
        return int(unpack("I", self.f.read(4))[0])

    def read_u64(self):
        """
        :rtype: int
        """
        return int(unpack("q", self.f.read(8))[0])

    def read_char(self):
        """
        :rtype: int
        """
        return unpack("b", self.f.read(1))[0]

    def read_str(self, length, encoding="ascii"):
        """
        :param int length:
        :param str encoding:
        :rtype: str
        """
        a = array.array("b")
        a.fromfile(self.f, length)
        return a.tostring().decode(encoding)

    def read_f32(self):
        """
        :rtype: float
        """
        return float(unpack("f", self.f.read(4))[0])

    def read_f64(self):
        """
        :rtype: float
        """
        return float(unpack("d", self.f.read(8))[0])

    def read_v(self, size, a):
        """
        :param int size:
        :param array.array a:
        :rtype: array.array
        """
        a.fromfile(self.f, size)
        return a

    def __init__(self, filename):
        """
        :param str filename:
        """
        self.header = "MIXSET\0"
        self.f = open(filename, "rb")
        header = self.read_str(8)
        assert header[:7] == self.header
        self.version = self.read_u32()
        self.dim = self.read_u32()

        self.num_means = self.read_u32()
        self.means = numpy.zeros([self.num_means, self.dim], dtype=numpy.float64)
        self.mean_weights = numpy.zeros(self.num_means, dtype=numpy.float64)

        for n in range(self.num_means):
            size = self.read_u32()
            assert size == self.dim
            arr_f = array.array("d")
            arr_f.fromfile(self.f, self.dim)
            self.means[n, :] = numpy.array(arr_f)
            self.mean_weights[n] = self.read_f64()

        self.num_covs = self.read_u32()
        self.covs = numpy.zeros([self.num_covs, self.dim], dtype=numpy.float64)
        self.cov_weights = numpy.zeros(self.num_covs, dtype=numpy.float64)

        for n in range(self.num_covs):
            size = self.read_u32()
            assert size == self.dim
            arr_f = array.array("d")
            arr_f.fromfile(self.f, self.dim)
            self.covs[n, :] = numpy.array(arr_f)
            self.cov_weights[n] = self.read_f64()

        self.num_densities = self.read_u32()
        self.densities = numpy.zeros((self.num_densities, 2), dtype=numpy.int32)
        for n in range(self.num_densities):
            mean_idx = self.read_u32()
            cov_idx = self.read_u32()
            self.densities[n, 0] = mean_idx
            self.densities[n, 1] = cov_idx

        self.num_mixtures = self.read_u32()
        self.mixtures = [
            None
        ] * self.num_mixtures  # type: typing.List[typing.Optional[typing.Tuple[typing.List[int],typing.List[float]]]]  # nopep8
        for n in range(self.num_mixtures):
            num_densities = self.read_u32()
            dns_idx = []
            dns_weight = []
            for i in range(num_densities):
                dns_idx.append(self.read_u32())
                dns_weight.append(self.read_f64())
            self.mixtures[n] = (dns_idx, dns_weight)

    # TODO?
    # noinspection PyUnresolvedReferences
    def write(self, filename):
        """
        :param str filename:
        """
        self.f = open(filename, "wb")
        self.write_str(self.header + "t")
        self.write_u32(self.version)
        self.write_u32(self.dim)

        self.write_u32(self.num_means)

        for n in range(self.num_means):
            self.write_u32(self.dim)
            arr_f = array.array("d")
            arr_f.fromlist(list(self.means[n, :]))
            arr_f.tofile(self.f)
            self.write_f64(self.mean_weights[n])

        self.write_u32(self.num_covs)

        for n in range(self.num_covs):
            self.write_u32(self.dim)
            arr_f = array.array("d")
            arr_f.fromlist(list(self.covs[n, :]))
            arr_f.tofile(self.f)
            self.write_f64(self.cov_weights[n])

        self.write_u32(self.num_densities)
        for n in range(self.num_densities):
            self.write_u32(int(self.densities[n, 0]))
            self.write_u32(int(self.densities[n, 1]))

        self.write_u32(self.num_mixtures)
        for n in range(self.num_mixtures):
            num_densities = len(self.mixtures[n][0])
            self.write_u32(num_densities)
            for i in range(num_densities):
                self.write_u32(self.mixtures[n][0][i])
                self.write_f64(self.mixtures[n][1][i])

    def get_mean_by_idx(self, idx):
        """
        :param int idx:
        :rtype: numpy.ndarray
        """
        return numpy.float32(self.means[idx, :] / self.mean_weights[idx])

    def get_cov_by_idx(self, idx):
        """
        :param int idx:
        :rtype: numpy.ndarray
        """
        return numpy.float32(self.covs[idx, :] / self.cov_weights[idx])

    def get_number_mixtures(self):
        """
        :rtype: int
        """
        return self.num_mixtures


class WordBoundaries:
    """
    Word boundaries.
    """

    # read routines
    def read_u16(self):
        """
        :rtype: int
        """
        return int(unpack("H", self.f.read(2))[0])

    def read_u32(self):
        """
        :rtype: int
        """
        return int(unpack("i", self.f.read(4))[0])

    def read_str(self, length, encoding="ascii"):
        """
        :param int length:
        :param str encoding:
        :rtype: str
        """
        a = array.array("b")
        a.fromfile(self.f, length)
        return a.tostring().decode(encoding)

    def __init__(self, filename):
        """
        :param str filename:
        """
        self.header = "LATWRDBN"
        self.f = open(filename, "rb")
        header = self.read_str(8)
        assert header == self.header
        self.version = self.read_u32()
        self.size = self.read_u32()
        print("version=%d size=%d" % (self.version, self.size))
        self.boundaries = []
        for i in range(self.size):
            time = self.read_u32()
            final = self.read_u16()
            initial = self.read_u16()
            bnd = (time, initial, final)
            self.boundaries.append(bnd)
            print(bnd)


###############################################################################


def main():
    """
    Main entry for usage as a tool.
    """
    from argparse import ArgumentParser

    arg_parser = ArgumentParser()
    arg_parser.add_argument("archive")
    arg_parser.add_argument("file", nargs="?")
    arg_parser.add_argument("--mode", default="list", help="list, show")
    arg_parser.add_argument("--type", default="feat", help="ascii, feat, align, bin-matrix, flow-cache")
    arg_parser.add_argument("--allophone-file")
    args = arg_parser.parse_args()

    a = open_file_archive(args.archive, must_exists=True)

    if args.mode == "list":
        print("\n".join(sorted(s for s in a.file_list())))
        sys.exit(0)

    elif args.mode == "show":
        assert args.file, "need to provide 'file' for --mode show. see --help"
        if args.type == "align":
            if args.allophone_file:
                a.set_allophones(args.allophone_file)

            f = a.read(args.file, "align")
            for time, index, state, weight in f:
                # Keep similar format as Sprint archiver.
                items = [
                    f"time={time}",
                    f"allophone={a.get_allophones_list()[index]}",
                    f"index={index}",
                    f"state={state}",
                ]
                if weight != 1:
                    items.append(f"weight={weight}")
                print("\t".join(items))

        elif args.type == "feat":
            t, f = a.read(args.file, "feat")
            for row, time in zip(f, t):
                print(str(time) + "--------" + " ".join("%.6f " % x for x in row))

        else:
            raise NotImplementedError("invalid --type %r" % args.type)

    else:
        raise NotImplementedError("invalid --mode %r" % args.mode)


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    main()



"""
Parametrize some parameters, e.g. to implement weight dropout, variational noise, weight norm, etc.

We follow the `PyTorch parametrization API
<https://pytorch.org/docs/stable/generated/torch.nn.utils.parametrize.register_parametrization.html>`__
and also borrow some code.

https://github.com/rwth-i6/returnn/issues/1518
"""

from __future__ import annotations
from typing import Optional, Union, Protocol
import copyreg
import weakref
import returnn.frontend as rf
from returnn.tensor import Tensor


__all__ = ["register_parametrization", "remove_parametrization", "is_parametrized"]


def register_parametrization(
    module: rf.Module, param_name: str, parametrization: _ParametrizationType, *, keep_existing_param: bool = True
) -> rf.Module:
    """
    Register parametrization for a tensor (parameter) in a module.

    :param module:
    :param param_name:
    :param parametrization:
    :param keep_existing_param:
        True: the original parameter stays in there,
            and parametrization will be called with the original parameter as an argument::
                parametrization(orig_param)
            In this case, parametrization must not have own parameters.
            This is useful for potential optional transformations, e.g. weight dropout or variational noise.
        False: the original parameter will be removed, and this will be a submodule,
            which can have its own parameters.
            It will be called without arguments::
                parametrization()
    """
    if not is_parametrized(module):
        # Sets up a module to be parametrized.
        # This works by substituting the class of the module by a class
        # that extends it to be able to inject a property.
        # We need this because we cannot inject a property into an object instance
        # (see https://docs.python.org/3/howto/descriptor.html)
        # and also we do not want to modify the original class.
        cls = module.__class__
        param_cls = _Metaclass(f"Parametrized{cls.__name__}", (cls,), {})
        module.__class__ = param_cls

    if hasattr(module.__class__, param_name):
        raise ValueError(
            f"register_parametrization: parametrized property {param_name} already exists in module {module}"
        )

    orig_param = getattr(module, param_name)
    if not isinstance(orig_param, rf.Parameter):
        raise TypeError(f"module.{param_name} is not a parameter, got {orig_param!r}")

    if keep_existing_param:
        if isinstance(parametrization, rf.Module):
            if len(list(parametrization.parameters())) > 0:
                raise ValueError(
                    f"register_parametrization: parametrization {parametrization} must not have parameters"
                    f" with keep_existing_param=True"
                )

    else:
        if hasattr(parametrization, "assign"):
            parametrization.assign(orig_param)
        orig_param = None
        # Put the parametrization into the module as a submodule
        # instead of the original parameter.
        # module.named_parameters() will thus find it, even when we install the new property.
        setattr(module, param_name, parametrization)

    # Injects a property into module
    assert isinstance(module.__class__, _Metaclass), "module must be parametrized"
    assert not hasattr(module.__class__, param_name), "property already exists"
    prop = _Property(module, param_name, parametrization, orig_param)
    setattr(module.__class__, param_name, prop)

    return module


def remove_parametrization(module: rf.Module, param_name: str) -> rf.Module:
    """
    Remove parametrization for a tensor (parameter) in a module.
    """
    if not is_parametrized(module):
        raise ValueError(f"module {module} is not parametrized")
    prop = getattr(module.__class__, param_name)
    assert isinstance(prop, _Property)
    delattr(module.__class__, param_name)
    assert not hasattr(module.__class__, param_name)
    if prop.orig_param is None:
        setattr(module, param_name, rf.Parameter(prop.parametrization()))
    # Check if there are any other parametrizations left.
    for k, v in vars(module.__class__).items():
        if isinstance(v, _Property):
            break
    else:  # no break, no other parametrizations
        module.__class__ = module.__class__.__bases__[0]  # revert to original class
    return module


def is_parametrized(module: rf.Module, param_name: Optional[str] = None) -> bool:
    r"""Returns ``True`` if module has an active parametrization.

    If the argument :attr:`tensor_name` is specified, returns ``True`` if
    ``module[tensor_name]`` is parametrized.

    Args:
        module: module to query
        param_name: attribute in the module to query
            Default: ``None``
    """
    if module.__class__.__class__ is not _Metaclass:
        return False
    if param_name is None:
        return True
    return hasattr(module.__class__, param_name)


class _ParametrizationTransform(Protocol):
    def __call__(self, x: Tensor) -> Tensor:
        """Return the parametrized tensor based on the original parameter."""


class _ParametrizationWithAssign(Protocol):
    def __call__(self) -> Tensor:
        """Return the parametrized tensor."""

    def assign(self, x: Tensor):
        """Assign as if it was a single parameter."""


class _ParametrizationWithoutAssign(Protocol):
    def __call__(self) -> Tensor:
        """Return the parametrized tensor."""


_ParametrizationType = Union[
    _ParametrizationTransform,
    _ParametrizationWithAssign,
    _ParametrizationWithoutAssign,
]


class _Metaclass(type):
    """
    https://stackoverflow.com/a/75943813/133374
    """


def _reduce_metaclass(cls):
    metaclass = cls.__class__
    cls_vars = dict(vars(cls))
    cls_vars.pop("__dict__", None)
    cls_vars.pop("__weakref__", None)
    return metaclass, (cls.__name__, cls.__bases__, cls_vars)


copyreg.pickle(_Metaclass, _reduce_metaclass)


class _Property:
    def __init__(
        self,
        module: rf.Module,
        param_name: str,
        parametrization: _ParametrizationType,
        orig_param: Optional[rf.Parameter],
    ):
        self.module_ref = weakref.ref(module)
        self.param_name = param_name
        self.parametrization = parametrization
        self.orig_param = orig_param

    def __get__(self, obj, objtype=None):
        if obj is None:  # called on the class
            return self
        assert obj is self.module_ref(), f"parametrize _Property __get__: {obj!r} vs {self.module_ref()!r}"
        if self.orig_param is not None:
            return self.parametrization(self.orig_param)
        else:
            return self.parametrization()

    def __set__(self, obj, value):
        assert obj is self.module_ref(), f"parametrize _Property __set__: {obj!r} vs {self.module_ref()!r}"
        if self.orig_param is not None:
            self.orig_param.assign(value)
        else:
            if hasattr(self.parametrization, "assign"):
                self.parametrization.assign(value)
            else:
                raise AttributeError(f"Cannot assign to {self.param_name} parametrization {self.parametrization}")



"""
From raw dict to extern_data tensor dict.
"""

from __future__ import annotations
from typing import Optional, Any, Union, Dict, List, Sequence
import numpy
import torch
from returnn.tensor import Tensor, TensorDict, Dim


def extern_data_template_from_config_opts(extern_data_dict: Dict[str, Any]) -> TensorDict:
    """
    :param extern_data_dict: as you would specify in the config
    :return: extern data tensor dict
    """
    extern_data = TensorDict()
    extern_data.update(extern_data_dict, auto_convert=True)
    if "seq_tag" not in extern_data.data:
        batch_dim = get_batch_dim_from_extern_data(extern_data)
        extern_data.data["seq_tag"] = Tensor(name="seq_tag", dtype="string", dims=[batch_dim])
    return extern_data


def raw_dict_to_extern_data(
    extern_data_raw: Dict[str, Union[torch.Tensor, numpy.ndarray]],
    *,
    extern_data_template: TensorDict,
    device: Union[str, torch.device],
    float_dtype: Optional[Union[str, torch.dtype]] = None,
) -> TensorDict:
    """
    :param extern_data_raw: This comes out of the DataLoader, via our collate_batch.
    :param extern_data_template: Specified via `extern_data` in the config.
    :param device: E.g. the GPU.
    :param float_dtype:
    :return: tensor dict, like extern_data_template, but with raw tensors set to Torch tensors, on the right device.
    """
    if isinstance(float_dtype, str):
        float_dtype = getattr(torch, float_dtype)
        assert isinstance(float_dtype, torch.dtype)
    assert isinstance(extern_data_raw, dict) and extern_data_raw
    batch_dim = get_batch_dim_from_extern_data(extern_data_template)
    for dim in _get_dyn_dims_from_extern_data(extern_data_template):
        dim.reset_eager()  # they will be reset below
    if batch_dim.size is None and batch_dim.dyn_size_ext is None:
        batch_dim.dyn_size_ext = Tensor(batch_dim.name or "batch", dims=[], dtype="int32")
    extern_data = TensorDict()
    for k, data in extern_data_template.data.items():
        data = data.copy_template()
        raw_tensor = extern_data_raw[k]
        assert len(raw_tensor.shape) == data.batch_ndim, f"ndim mismatch for {k}: {raw_tensor.shape} vs {data}"
        for i, dim in enumerate(data.dims):
            if dim.dimension is not None:
                assert (
                    dim.dimension == raw_tensor.shape[i]
                ), f"shape mismatch for {k}: {raw_tensor.shape} vs {data.batch_shape}"
        if isinstance(raw_tensor, torch.Tensor):
            if raw_tensor.dtype.is_floating_point and float_dtype:
                raw_tensor = raw_tensor.to(dtype=float_dtype)
            data.dtype = str(raw_tensor.dtype).split(".")[-1]  # just overwrite for now...
            data.raw_tensor = raw_tensor.to(device)
        elif isinstance(raw_tensor, numpy.ndarray):
            data.raw_tensor = raw_tensor  # leave it as it is
        else:
            raise TypeError(f"Unexpected type {type(raw_tensor)} for {k} in extern_data_raw.")

        if batch_dim.dyn_size_ext is not None and batch_dim.dyn_size_ext.raw_tensor is None:
            batch_dim.dyn_size_ext.raw_tensor = torch.tensor(extern_data_raw[k].shape[0], dtype=torch.int32)

        # This has certain assumptions on the dataset, the data pipeline and collate_batch.
        # Namely, we expect that we get the batch dim in the first dim (see collate_batch).
        # We also expect that the sequence lengths are in the second dim, if it is dynamic.
        if (
            len(data.dims) >= 2
            and data.dims[1].size is None
            and (data.dims[1].dyn_size_ext is None or data.dims[1].dyn_size_ext.raw_tensor is None)
        ):
            assert k + ":seq_len" in extern_data_raw, (
                f"extern_data {data}, dyn spatial dim, missing {k}:seq_len in raw dict, "
                f"check dataset or collate_batch"
            )
            size = extern_data_raw[k + ":seq_len"]
            # Sequence lengths have to be on CPU for the later call to rnn.pack_padded_sequence
            assert size.device.type == "cpu"
            size_dtype = str(size.dtype).split(".")[-1]
            if data.dims[1].dyn_size_ext is None:
                data.dims[1].dyn_size_ext = Tensor(data.dims[1].name or "time", dims=[batch_dim], dtype=size_dtype)
            data.dims[1].dyn_size_ext.dtype = size_dtype
            data.dims[1].dyn_size_ext.raw_tensor = size

        extern_data.data[k] = data

    return extern_data


def raw_dict_can_split_batch(
    extern_data_raw: Dict[str, Union[torch.Tensor, numpy.ndarray]],
    *,
    num_splits: int = 2,
) -> bool:
    """
    :param extern_data_raw: This comes out of the DataLoader, via our collate_batch.
    :param num_splits: into how many parts
    :return:
    """
    some_value = next(iter(extern_data_raw.values()))
    assert isinstance(some_value, (torch.Tensor, numpy.ndarray))
    batch_size = int(some_value.shape[0])
    return batch_size >= num_splits


def raw_dict_split_batch(
    extern_data_raw: Dict[str, Union[torch.Tensor, numpy.ndarray]],
    *,
    splits: Union[Sequence[int], int],
) -> List[Dict[str, Union[torch.Tensor, numpy.ndarray]]]:
    """
    :param extern_data_raw: This comes out of the DataLoader, via our collate_batch.
    :param splits: either the list of resulting batch dims, or the num of splits
    :return: splitted extern_data_raw
    """
    some_value = next(iter(extern_data_raw.values()))
    assert isinstance(some_value, (torch.Tensor, numpy.ndarray))
    batch_size = int(some_value.shape[0])
    if isinstance(splits, int):
        splits = _make_split_seq_from_num_splits(splits, batch_size)
    assert isinstance(splits, (tuple, list)) and sum(splits) == batch_size
    res: List[Dict[str, Union[torch.Tensor, numpy.ndarray]]] = [{} for _ in range(len(splits))]
    for k, v in extern_data_raw.items():
        if not isinstance(v, (torch.Tensor, numpy.ndarray)):
            raise TypeError(f"got invalid value of type ({type(v).__name__}) for key {k!r}")
        offset = 0
        for i, split_size in enumerate(splits):
            res[i][k] = v[offset : offset + split_size] if v.ndim > 0 else v
            offset += split_size
    for res_ in res:
        res_: Dict[str, Union[torch.Tensor, numpy.ndarray]]
        for k, v in res_.items():
            if "%s:seq_len" % k in res_:
                max_len = torch.max(res_["%s:seq_len" % k])
                res_[k] = v[:, :max_len]
    return res


def _make_split_seq_from_num_splits(num_splits: int, total: int) -> List[int]:
    assert num_splits <= total
    remaining = total
    split_size = -(-total // num_splits)  # ceildiv
    res = []
    for i in range(num_splits):
        res.append(min(remaining, split_size))
        remaining -= res[-1]
    assert remaining == 0 and len(res) == num_splits and sum(res) == total
    return res


def _get_dyn_dims_from_extern_data(extern_data: TensorDict) -> List[Dim]:
    visited = set()
    res = []
    for k, v in extern_data.data.items():
        for dim in v.dims:
            if dim not in visited and dim.size is None:
                visited.add(dim)
                res.append(dim)
    return res


def get_batch_dim_from_extern_data(extern_data: TensorDict) -> Dim:
    """
    We expect that the batch dim is the first dim in any of the tensors.
    See collate_batch.

    We allow that the batch dim is not necessarily the global batch_dim object.
    We also allow that this is not even marked as batch dim (is_batch_dim() can be False).
    """
    batch_dim = next(iter(extern_data.data.values())).dims[0]
    return batch_dim
